import contextvars
import json
import typing
from contextlib import contextmanager

import pyarrow as pa

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import (
    arrow_util,
    artifact_base,
    box,
    mappers,
    mappers_weave,
    node_ref,
    partial_object,
    ref_base,
)
from weave.legacy.weave import mappers_python_def as mappers_python
from weave.legacy.weave.arrow import arrow
from weave.legacy.weave.language_features.tagging import tagged_value_type

_in_tagging_context = contextvars.ContextVar("in_tagging_context", default=False)


@contextmanager
def _strings_as_dictionaries():
    token = _in_tagging_context.set(True)
    try:
        yield
    finally:
        _in_tagging_context.reset(token)


class TypedDictToArrowStruct(mappers_python.TypedDictToPyDict):
    def result_type(self):
        fields = []
        for property_key, property_serializer in self._property_serializers.items():
            prop_result_type = property_serializer.result_type()
            fields.append(arrow_util.arrow_field(property_key, prop_result_type))
        return pa.struct(fields)


class ObjectToArrowStruct(mappers_python.ObjectToPyDict):
    def result_type(self):
        fields = []
        for property_key, property_serializer in self._property_serializers.items():
            if property_serializer is not None:
                prop_result_type = property_serializer.result_type()
                fields.append(arrow_util.arrow_field(property_key, prop_result_type))
        return pa.struct(fields)


class TimeDeltaToArrowTimeDelta(mappers.Mapper):
    def result_type(self):
        return pa.duration("us")


class ArrowTimeDeltaToTimeDelta(mappers.Mapper):
    def apply(self, obj):
        return obj.as_py()


class StringToArrow(mappers_python.StringToPyString):
    def result_type(self):
        if _in_tagging_context.get():
            return pa.dictionary(pa.int64(), pa.string(), False)
        return pa.string()


class TaggedValueToArrowStruct(tagged_value_type.TaggedValueToPy):
    def result_type(self):
        with _strings_as_dictionaries():
            tag_type = self._tag_serializer.result_type()
        value_type = self._value_serializer.result_type()

        return pa.struct(
            [
                arrow_util.arrow_field("_tag", tag_type),
                arrow_util.arrow_field("_value", value_type),
            ]
        )


class ListToArrowArr(mappers_python.ListToPyList):
    def result_type(self):
        # It is important to use `item` here since pyarrow auto assigns the
        # inner field name of a list to `item`. Anything else can cause issues
        # with concatenation or other pyarrow operations that need type
        # equality.
        return pa.list_(arrow_util.arrow_field("item", self._object_type.result_type()))


class UnionToArrowUnion(mappers_weave.UnionMapper):
    def __init__(self, type_, mapper, artifact, path=[]):
        super().__init__(type_, mapper, artifact, path)
        if not self.is_single_object_nullable:
            self._type_codes: dict[types.Type, int] = {}
            self._type_code_to_member_mapper: dict[int, mappers.Mapper] = {}

            i: int = 0
            for member_mapper in self._member_mappers:
                member_type = member_mapper.type
                if isinstance(member_type, types.NoneType):
                    continue
                self._type_codes[member_type] = i
                self._type_code_to_member_mapper[i] = member_mapper
                i += 1

    @property
    def non_none_members(self):
        return [
            m for m in self._member_mappers if not isinstance(m.type, types.NoneType)
        ]

    def result_type(self):
        nullable = False
        non_null_mappers: typing.List[mappers.Mapper] = []
        for member_mapper in self._member_mappers:
            if isinstance(member_mapper.type, types.NoneType):
                nullable = True
            else:
                non_null_mappers.append(member_mapper)
        if not nullable or len(non_null_mappers) > 1:
            fields = [
                pa.field(
                    str(i),
                    non_null_mappers[i].result_type(),
                    nullable,
                )
                for i in range(len(non_null_mappers))
            ]
            return pa.union(fields, mode="dense")
        return arrow_util.arrow_type_with_nullable(non_null_mappers[0].result_type())

    def type_code_of_type(self, type: types.Type) -> int:
        """Return the arrow type code for the given type."""
        if self.is_single_object_nullable:
            raise errors.WeaveTypeError(
                "Cannot get type code of type in a union that is nullable"
            )
        elif isinstance(type, types.NoneType):
            raise errors.WeaveTypeError("Cannot get type code of NoneType")
        else:
            try:
                return self._type_codes[type]
            except KeyError:
                for member_type, member_type_code in self._type_codes.items():
                    if not isinstance(
                        types.merge_types(member_type, type), types.UnionType
                    ):
                        return member_type_code
                raise errors.WeaveTypeError(f"Could not find type code for {type}")

    def type_code_of_obj(self, obj) -> int:
        """Return the arrow type code for the given object."""
        obj_type = types.TypeRegistry.type_of(obj)
        return self.type_code_of_type(obj_type)

    def mapper_of_type_code(self, type_code: int) -> mappers.Mapper:
        """Return the mapper for the given type code."""
        if self.is_single_object_nullable:
            raise errors.WeaveTypeError(
                "Cannot get mapper of type code in a union that is nullable"
            )
        try:
            return self._type_code_to_member_mapper[type_code]
        except KeyError:
            raise errors.WeaveTypeError(
                f"Could not find mapper for type code {type_code}"
            )

    def apply(self, obj):
        obj_type = types.TypeRegistry.type_of(obj)
        for member_mapper in self._member_mappers:
            if member_mapper.type.assign_type(obj_type):
                return member_mapper.apply(obj)
        raise errors.WeaveTypeError(f"Could not find type for {obj}")


class ArrowUnionToUnion(UnionToArrowUnion):
    def apply(self, obj):
        if self.is_single_object_nullable:
            if obj is None:
                return None
            for mapper in self._member_mappers:
                if not isinstance(mapper.type, types.NoneType):
                    return mapper.apply(obj)
        if not isinstance(obj, pa.UnionScalar):
            raise errors.WeaveTypeError(f"Expected UnionScalar, got {type(obj)}: {obj}")
        mapper = self._member_mappers[obj.type_code]
        return mapper.apply(obj.as_py())


class IntToArrowInt(mappers_python.IntToPyInt):
    def result_type(self):
        return pa.int64()


class BoolToArrowBool(mappers_python.BoolToPyBool):
    def result_type(self):
        return pa.bool_()


class TimestampToArrowTimestamp(mappers_python.TimestampToPyTimestamp):
    def result_type(self):
        return pa.timestamp("ms", tz="UTC")


class FloatToArrowFloat(mappers.Mapper):
    def result_type(self):
        return pa.float64()

    def apply(self, obj):
        return obj


class ArrowFloatToFloat(mappers.Mapper):
    def apply(self, obj):
        return obj


class ArrowStringToString(mappers.Mapper):
    def apply(self, obj):
        return obj


class FunctionToArrowFunction(mappers.Mapper):
    def result_type(self):
        return pa.string()

    def apply(self, obj):
        ref = node_ref.node_to_ref(obj)
        return str(ref)


class ArrowFunctionToFunction(mappers.Mapper):
    def apply(self, obj):
        ref = ref_base.Ref.from_str(obj)
        return node_ref.ref_to_node(ref)


class NoneToArrowNone(mappers.Mapper):
    def result_type(self):
        return pa.null()


class UnknownToArrowNone(mappers_python.UnknownToPyUnknown):
    def result_type(self):
        return pa.null()


class ArrowDateTimeToDateTime(mappers.Mapper):
    def apply(self, obj):
        return obj


class DefaultToArrow(mappers_python.DefaultToPy):
    def result_type(self):
        # TODO: hard-coding for the moment. Need to generalize this.
        # There are two cases, either there's a instance_to_dict() method
        #     in which case we need to know the types of that dict
        #     (this is similar to ObjectType and should be shared somehow).
        # Or we'll use save_instance which return a RefType (which we encode
        #     as a pyarrow string).

        if self.type.name == "run":
            return pa.struct(
                (
                    pa.field("entity_name", pa.string()),
                    pa.field("project_name", pa.string()),
                    pa.field("run_id", pa.string()),
                )
            )
        elif self.type.name == "artifact":
            return pa.struct(
                (
                    pa.field("entity_name", pa.string()),
                    pa.field("project_name", pa.string()),
                    pa.field("artifact_type_name", pa.string()),
                    pa.field("artifact_name", pa.string()),
                )
            )
        elif self.type.name == "panel":
            return pa.struct(
                (
                    pa.field("id", pa.string()),
                    pa.field("input", pa.string()),
                    pa.field("config", pa.string()),
                )
            )
        elif (
            self.type.name == "WeaveNDArray"
            or self.type.name == "pil_image"
            or self.type.name == "ArrowArray"
            or self.type.name == "ArrowTable"
            or self.type.name == "FilesystemArtifact"
            or self.type.name == "file"
            or self.type.name == "WandbArtifactRef"
            or self.type.name == "LocalArtifactRef"
            or self.type.name == "OpDef"
        ):
            # Ref type
            return pa.string()
        elif self.type.name == "timestamp":
            return pa.timestamp("ms", tz="+00:00")
        elif self.type.name == "date":
            return pa.timestamp("ms", tz="+00:00")
        elif self.type.name == "type":
            return pa.string()
        elif self.type.name == "ndarray":
            return pa.null()

        raise errors.WeaveInternalError(
            "Type not yet handled by mappers_arrow: %s" % self.type
        )

    def apply(self, obj):
        # Hacking... when its a panel, we need to json dump the node and config
        res = super().apply(obj)
        if self.type.name == "panel":
            res["input"] = json.dumps(res["input"])
            res["config"] = json.dumps(res["config"])
        return res


class DefaultFromArrow(mappers_python.DefaultFromPy):
    def apply(self, obj):
        # Hacking... when its a panel, we need to json load the node and config
        if self.type.name == "panel":
            obj["input"] = json.loads(obj["input"])
            obj["config"] = json.loads(obj["config"])
        return super().apply(obj)


class ArrowToArrowWeaveListOrPylist(mappers_python.ListToPyList):
    def apply(self, obj):
        if isinstance(self.type, arrow.ArrowWeaveListType):
            # we're already mapped - no need to go further
            return obj

        return super().apply(obj)


class GQLHasKeysToArrowStruct(mappers_python.GQLClassWithKeysToPyDict):
    def result_type(self):
        return pa.struct(
            [
                pa.field(key, self._property_serializers[key].result_type())
                for key in self.type.keys
            ]
        )

    def as_typeddict_mapper(self):
        typed_dict_type = types.TypedDict(self.type.keys)
        return TypedDictToArrowStruct(
            typed_dict_type, self.mapper, self._artifact, self.path
        )


def map_to_arrow_(
    type, mapper, artifact: artifact_base.Artifact, path=[], mapper_options=None
):
    if isinstance(type, types.Const):
        type = type.val_type
    if isinstance(type, types.TypedDict):
        return TypedDictToArrowStruct(type, mapper, artifact, path)
    elif isinstance(type, types.List) or isinstance(type, arrow.ArrowWeaveListType):
        return ListToArrowArr(type, mapper, artifact, path)
    elif isinstance(type, types.UnionType):
        return UnionToArrowUnion(type, mapper, artifact, path)
    elif isinstance(type, partial_object.PartialObjectType):
        return GQLHasKeysToArrowStruct(type, mapper, artifact, path)
    elif isinstance(type, types.ObjectType):
        return ObjectToArrowStruct(type, mapper, artifact, path)
    elif isinstance(type, tagged_value_type.TaggedValueType):
        return TaggedValueToArrowStruct(type, mapper, artifact, path)  # type: ignore
    elif isinstance(type, types.Int):
        return IntToArrowInt(type, mapper, artifact, path)
    elif isinstance(type, types.Boolean):
        return BoolToArrowBool(type, mapper, artifact, path)
    elif isinstance(type, types.Float) or isinstance(type, types.Number):
        return FloatToArrowFloat(type, mapper, artifact, path)
    elif isinstance(type, types.String):
        return StringToArrow(type, mapper, artifact, path)
    elif isinstance(type, types.Timestamp):
        return TimestampToArrowTimestamp(type, mapper, artifact, path)
    elif isinstance(type, types.Function):
        return FunctionToArrowFunction(type, mapper, artifact, path)
    elif isinstance(type, types.NoneType):
        return NoneToArrowNone(type, mapper, artifact, path)
    elif isinstance(type, types.UnknownType):
        return UnknownToArrowNone(type, mapper, artifact, path)
    elif isinstance(type, types.TimeDelta):
        return TimeDeltaToArrowTimeDelta(type, mapper, artifact, path)
    else:
        return DefaultToArrow(type, mapper, artifact, path)


def map_from_arrow_(type, mapper, artifact, path=[], mapper_options=None):
    if isinstance(type, types.TypedDict):
        return mappers_python.TypedDictToPyDict(type, mapper, artifact, path)
    elif isinstance(type, (types.List, arrow.ArrowWeaveListType)):
        return ArrowToArrowWeaveListOrPylist(type, mapper, artifact, path)
    elif isinstance(type, types.UnionType):
        return ArrowUnionToUnion(type, mapper, artifact, path)
    elif isinstance(type, partial_object.PartialObjectType):
        return mappers_python.GQLClassWithKeysToPyDict(type, mapper, artifact, path)
    elif isinstance(type, types.ObjectType):
        return mappers_python.ObjectDictToObject(type, mapper, artifact, path)
    elif isinstance(type, tagged_value_type.TaggedValueType):
        return tagged_value_type.TaggedValueFromPy(type, mapper, artifact, path)
    elif isinstance(type, types.Int):
        return mappers_python.IntToPyInt(type, mapper, artifact, path)
    elif isinstance(type, types.Boolean):
        return mappers_python.BoolToPyBool(type, mapper, artifact, path)
    elif isinstance(type, types.Float) or isinstance(type, types.Number):
        return ArrowFloatToFloat(type, mapper, artifact, path)
    elif isinstance(type, types.String):
        return ArrowStringToString(type, mapper, artifact, path)
    elif isinstance(type, types.Timestamp):
        return ArrowDateTimeToDateTime(type, mapper, artifact, path)
    elif isinstance(type, types.Function):
        return ArrowFunctionToFunction(type, mapper, artifact, path)
    elif isinstance(type, types.NoneType):
        return mappers_python.NoneToPyNone(type, mapper, artifact, path)
    elif isinstance(type, types.UnknownType):
        return UnknownToArrowNone(type, mapper, artifact, path)
    elif isinstance(type, types.TimeDelta):
        return ArrowTimeDeltaToTimeDelta(type, mapper, artifact, path)
    else:
        return DefaultFromArrow(type, mapper, artifact, path)


map_to_arrow = mappers.make_mapper(map_to_arrow_)
map_from_arrow = mappers.make_mapper(map_from_arrow_)



# Notes of stuff to do:
# - Add filter controls
# - Make groupby a dropdown
# - Better categorical detection. Currently we can't tell if something is a unique
#   string column or categorical. There are three options I see:
#   - put this information in the type system
#   - fully weavify the Render body here so it can switch based on the result of .execute
#     (execute a grouping operation on the input node to get summary stats, then switch
#     behavior based on the result, or make a summary_stats op and do the same).
#   - allow panel render bodies to call use. The problem here is we don't want rendering
#     to incur round-trips. Maybe its already ok since we only recall python panel
#     renders when the input node or type changes, but no other time.
#   - There might be a more appealing edge based alternative... like if we could use
#     .initialize to do the summary, and stick that information in the panel's config?
#     that might work!
# - Multiple segments (allow defining multiple segment filters, concat tables together
#   including segment name, and then groupby segment name)
# - An earlier version of this used section layout instead of grid, and used ChildPanel
#   vars instead of VarBar, and it looked really nice in a notebook. We can make it
#   so you can switch between configurable panel (that can be added into an existing
#   Board), v whole new board layout. The information is the same in both cases, we
#   just use different UI structures to render.


import os
import typing

import weave
from weave.legacy.weave import weave_internal
from weave.legacy.weave.panels import panel_plot
from weave.legacy.weave.panels_py.generator_templates import template_registry
from weave.legacy.weave import util


@weave.type()
class AutoBoardConfig:
    pass


# Was putting in these midpoint functions, but date-add and some other
# date functions don't have arrow equivalents yet, so these don't vectorize
# @weave.op(
#     weavify=True,
#     name="numberbin-midpoint",
#     input_type={"bin": weave.types.NumberBinType},
# )
# def number_bin_midpoint(bin) -> float:
#     return (bin["start"] + bin["stop"]) / 2


# @weave.op(
#     weavify=True,
#     name="timestampbin-midpoint",
#     input_type={"bin": weave.types.TimestampBinType},
# )
# def timestamp_bin_midpoint(bin) -> float:
#     return bin["start"].add((bin["stop"].sub(bin["start"])).div(2))


def timeseries(
    input_node: weave.Node[list[typing.Any]],
    bin_domain_node: weave.Node,
    x_axis_key: str,
    y_expr: typing.Callable,
    y_title: str,
    color_expr: typing.Callable,
    color_title: str,
    x_domain: weave.Node,
    n_bins: int,
    mark: panel_plot.MarkOption = "line",
) -> weave.Panel:
    x_axis_type = input_node[x_axis_key].type.object_type  # type: ignore
    if weave.types.optional(weave.types.Timestamp()).assign_type(x_axis_type):
        x_title = ""
        bin_fn = weave.legacy.weave.ops.timestamp_bins_nice
    elif weave.types.optional(weave.types.Number()).assign_type(x_axis_type):
        x_title = x_axis_key
        bin_fn = weave.legacy.weave.ops.numbers_bins_equal
    else:
        raise ValueError(f"Unsupported type for x_axis_key {x_axis_key}: {x_axis_type}")
    if mark == "bar":
        x = lambda row: row[x_axis_key].bin(bin_fn(bin_domain_node, n_bins))
    else:
        # TODO: should be midpoint
        x = lambda row: row[x_axis_key].bin(bin_fn(bin_domain_node, n_bins))["start"]

    return weave.legacy.weave.panels.Plot(
        input_node,
        x=x,
        x_title=x_title,
        y=y_expr,
        y_title=y_title,
        label=color_expr,
        color_title=color_title,
        groupby_dims=["x", "label"],
        mark=mark,
        domain_x=x_domain,
    )


def timeseries_avg_line(
    input_node: weave.Node[list[typing.Any]],
    bin_domain_node: weave.Node,
    x_axis_key: str,
    y_axis_key: str,
    groupby_key: typing.Union[weave.Node[str], str],
    x_domain: weave.Node,
) -> weave.Panel:
    x_axis_type = input_node[x_axis_key].type.object_type  # type: ignore
    if weave.types.optional(weave.types.Timestamp()).assign_type(x_axis_type):
        bin_fn = weave.legacy.weave.ops.timestamp_bins_nice
    elif weave.types.optional(weave.types.Number()).assign_type(x_axis_type):
        bin_fn = weave.legacy.weave.ops.numbers_bins_equal
    else:
        raise ValueError(f"Unsupported type for x_axis_key {x_axis_key}: {x_axis_type}")
    return weave.legacy.weave.panels.Plot(
        input_node,
        x=lambda row: row[x_axis_key].bin(bin_fn(bin_domain_node, 100))["start"],
        x_title=x_axis_key,
        y=lambda row: row[y_axis_key].avg(),
        y_title="avg" + y_axis_key,
        label=lambda row: row[groupby_key],
        groupby_dims=["x", "label"],
        mark="line",
        no_legend=True,
        domain_x=x_domain,
    )


def timeseries_sum_bar(
    input_node: weave.Node[list[typing.Any]],
    bin_domain_node: weave.Node,
    x_axis_key: str,
    y_axis_key: str,
    groupby_key: typing.Union[weave.Node[str], str],
    x_domain: weave.Node,
    n_bins: int,
) -> weave.Panel:
    x_axis_type = input_node[x_axis_key].type.object_type  # type: ignore
    if weave.types.optional(weave.types.Timestamp()).assign_type(x_axis_type):
        x_title = ""
        bin_fn = weave.legacy.weave.ops.timestamp_bins_nice
    elif weave.types.optional(weave.types.Number()).assign_type(x_axis_type):
        x_title = x_axis_key
        bin_fn = weave.legacy.weave.ops.numbers_bins_equal
    else:
        raise ValueError(f"Unsupported type for x_axis_key {x_axis_key}: {x_axis_type}")
    return weave.legacy.weave.panels.Plot(
        input_node,
        x=lambda row: row[x_axis_key].bin(bin_fn(bin_domain_node, n_bins)),
        x_title=x_title,
        y=lambda row: row[y_axis_key].sum(),
        y_title="sum_" + y_axis_key,
        label=lambda row: row[groupby_key],
        groupby_dims=["x", "label"],
        mark="bar",
        no_legend=True,
        domain_x=x_domain,
    )


def timeseries_count_bar(
    input_node: weave.Node[list[typing.Any]],
    bin_domain_node: weave.Node,
    x_axis_key: str,
    groupby_key: typing.Union[weave.Node[str], str],
    x_domain: weave.Node,
    n_bins: int,
) -> weave.Panel:
    x_axis_type = input_node[x_axis_key].type.object_type  # type: ignore
    if weave.types.optional(weave.types.Timestamp()).assign_type(x_axis_type):
        x_title = ""
        bin_fn = weave.legacy.weave.ops.timestamp_bins_nice
    elif weave.types.optional(weave.types.Number()).assign_type(x_axis_type):
        x_title = x_axis_key
        bin_fn = weave.legacy.weave.ops.numbers_bins_equal
    else:
        raise ValueError(f"Unsupported type for x_axis_key {x_axis_key}: {x_axis_type}")
    return weave.legacy.weave.panels.Plot(
        input_node,
        x=lambda row: row[x_axis_key].bin(bin_fn(bin_domain_node, n_bins)),
        x_title=x_title,
        y=lambda row: row.count(),
        label=lambda row: row[groupby_key],
        groupby_dims=["x", "label"],
        mark="bar",
        domain_x=x_domain,
    )


def categorical_dist(
    input_node: weave.Node[list[typing.Any]],
    key: str,
) -> weave.Panel:
    return weave.legacy.weave.panels.Plot(
        input_node,
        y=lambda row: row[key],
        x=lambda row: row.count(),
        label=lambda row: row[key],
        groupby_dims=["y", "label"],
        mark="bar",
        no_legend=True,
    )


def node_qualifies_for_autoboard(input_node: weave.Node[typing.Any]) -> bool:
    input_type = input_node.type
    if not weave.types.is_list_like(input_type):
        return False

    object_type = input_type.object_type  # type: ignore
    if not weave.types.TypedDict().assign_type(object_type):
        return False
    property_types = typing.cast(
        dict[str, weave.types.Type], object_type.property_types
    )
    x_axis = None

    for k, prop_type in property_types.items():
        if not weave.types.NoneType().assign_type(prop_type) and weave.types.optional(
            weave.types.Timestamp()
        ).assign_type(prop_type):
            x_axis = k
            x_axis_type = prop_type
            break
    if x_axis is None:
        for step_key in ["_step", "step"]:
            if step_key in property_types and weave.types.optional(
                weave.types.Int()
            ).assign_type(property_types[step_key]):
                x_axis = step_key
                x_axis_type = property_types[step_key]

    if x_axis is None:
        return False

    return True


def auto_panels(
    input_node: weave.Node[list[typing.Any]],
    config: typing.Optional[AutoBoardConfig] = None,
) -> weave.Panel:
    input_type = input_node.type
    if isinstance(input_type, weave.types.Function):
        input_type = input_type.output_type
    if not weave.types.is_list_like(input_type):
        raise ValueError("Input node must be a list")
    object_type = input_type.object_type  # type: ignore
    if not weave.types.TypedDict().assign_type(object_type):
        raise ValueError("Input node must be a list of objects")
    property_types = typing.cast(
        dict[str, weave.types.Type], object_type.property_types
    )
    x_axis = None

    for k, prop_type in property_types.items():
        if not weave.types.NoneType().assign_type(prop_type) and weave.types.optional(
            weave.types.Timestamp()
        ).assign_type(prop_type):
            x_axis = k
            x_axis_type = prop_type
            break
    if x_axis is None:
        for step_key in ["_step", "step"]:
            if step_key in property_types and weave.types.optional(
                weave.types.Int()
            ).assign_type(property_types[step_key]):
                x_axis = step_key
                x_axis_type = property_types[step_key]

    if x_axis is None:
        # TODO: handle instead
        raise ValueError(
            'No suitable x-axis (Timestamp type, or _step) property found in input type "%s"'
            % input_type
        )

    groupby = weave_internal.const(None)
    for k, prop_type in property_types.items():
        if ("version" in k or "snapshot" in k) and weave.types.optional(
            weave.types.String()
        ).assign_type(prop_type):
            groupby = weave_internal.const(k)

    metric_panels = []
    categorical_panels = []

    data_node = weave_internal.make_var_node(input_node.type, "data")
    x_domain_node = weave_internal.make_var_node(
        weave.types.optional(weave.types.List(x_axis_type)), "zoom_range"
    )
    time_domain_node = weave_internal.make_var_node(
        weave.types.List(x_axis_type), "bin_range"
    )
    window_data_node = weave_internal.make_var_node(input_node.type, "window_data")

    groupby_key_node = weave_internal.make_var_node(groupby.type, "groupby")

    for key, prop_type in property_types.items():
        if weave.types.optional(weave.types.Number()).assign_type(prop_type):
            panel = timeseries_sum_bar(
                data_node,
                time_domain_node,
                x_axis,
                key,
                groupby_key_node,
                x_domain_node,
                50,
            )
            metric_panels.append(panel)
        elif (
            weave.types.optional(weave.types.String()).assign_type(prop_type)
            and key != "prompt"
            and key != "completion"
        ):
            panel = categorical_dist(window_data_node, key)
            categorical_panels.append(panel)

    metrics = weave.legacy.weave.panels.Group(
        layoutMode=weave.legacy.weave.panels.GroupLayoutFlow(2, 3),
        items={"panel%s" % i: panel for i, panel in enumerate(metric_panels)},
    )
    categoricals = weave.legacy.weave.panels.Group(
        layoutMode=weave.legacy.weave.panels.GroupLayoutFlow(2, 3),
        items={"panel%s" % i: panel for i, panel in enumerate(categorical_panels)},
    )
    control_items = [
        weave.legacy.weave.panels.GroupPanel(
            input_node,
            id="data",
        ),
        # TODO: We need a filter editor. Can start with a filter expression
        # editor and make it more user-friendly later
        weave.legacy.weave.panels.GroupPanel(
            lambda data: weave.legacy.weave.ops.make_list(
                a=data[x_axis].min(), b=data[x_axis].max()
            ),
            id="data_range",
            hidden=True,
        ),
        weave.legacy.weave.panels.GroupPanel(None, id="zoom_range", hidden=True),
        weave.legacy.weave.panels.GroupPanel(
            lambda zoom_range, data_range: zoom_range.coalesce(data_range),
            id="bin_range",
            hidden=True,
        ),
        weave.legacy.weave.panels.GroupPanel(
            lambda data, zoom_range: weave.legacy.weave.panels.DateRange(
                zoom_range, domain=data[x_axis]
            ),
            id="date_picker",
        ),
        # TODO: groupby should really be a Dropdown / multi-select instead
        # of an expression
        weave.legacy.weave.panels.GroupPanel(
            groupby,
            id="groupby",
        ),
        weave.legacy.weave.panels.GroupPanel(
            lambda data, bin_range: data.filter(
                lambda row: weave.legacy.weave.ops.Boolean.bool_and(
                    row[x_axis] >= bin_range[0], row[x_axis] < bin_range[1]
                )
            ),
            id="window_data",
            hidden=True,
        ),
    ]

    panels = [
        weave.legacy.weave.panels.BoardPanel(
            metrics,
            id="metrics",
            layout=weave.legacy.weave.panels.BoardPanelLayout(x=0, y=0, w=24, h=12),
        ),
        weave.legacy.weave.panels.BoardPanel(
            categoricals,
            id="categoricals",
            layout=weave.legacy.weave.panels.BoardPanelLayout(x=0, y=0, w=24, h=12),
        ),
    ]
    if "step" not in x_axis:
        panels.insert(
            0,
            weave.legacy.weave.panels.BoardPanel(
                timeseries_count_bar(
                    data_node,
                    time_domain_node,
                    x_axis,
                    groupby_key_node,
                    x_domain_node,
                    150,
                ),
                id="volume",
                layout=weave.legacy.weave.panels.BoardPanelLayout(x=0, y=0, w=24, h=6),
            ),
        )

    panels.append(
        weave.legacy.weave.panels.BoardPanel(
            window_data_node,
            id="table",
            layout=weave.legacy.weave.panels.BoardPanelLayout(x=0, y=0, w=24, h=6),
        ),
    )
    return weave.legacy.weave.panels.Board(vars=control_items, panels=panels, editable=False)


# The interface for constructing this Panel from Python
@weave.type()
class AutoBoard(weave.Panel):
    id = "AutoBoard"
    input_node: weave.Node[list[typing.Any]]
    config: typing.Optional[AutoBoardConfig] = None

    @weave.op()  # type: ignore
    def render(self) -> weave.legacy.weave.panels.Group:
        return auto_panels(self.input_node, self.config)  # type: ignore


@weave.op(  # type: ignore
    name="py_board-seed_autoboard",
    hidden=True,
)
def seed_autoboard(
    input_node: weave.Node[typing.Any],
    config: typing.Optional[AutoBoardConfig] = None,
) -> weave.legacy.weave.panels.Group:
    return auto_panels(input_node, config)  # type: ignore


instructions_md = util.read_or_default(
    os.path.join(os.path.dirname(__file__), "instructions", "panel_autoboard.md")
)


# Removing this from registry for now since it is pretty buggy and not well tested.
# template_registry.register(
#     "py_board-seed_autoboard",
#     "Timeseries Auto-Board",
#     "Column-level analysis of timeseries data",
#     input_node_predicate=node_qualifies_for_autoboard,
#     # Not featuring because it is pretty buggy
#     # is_featured=True,
#     instructions_md=instructions_md,
# )



from typing import Any, Optional

import pydantic

import weave
from weave.legacy.weave import weave_types as types


def test_pydantic_type_inference():
    class TestPydanticClass(pydantic.BaseModel):
        a: int
        b: str
        c: float

    obj = TestPydanticClass(a=1, b="2", c=3.0)
    t = types.TypeRegistry.type_of(obj)
    assert isinstance(t, types.ObjectType)
    assert t.property_types() == {
        "a": types.Int(),
        "b": types.String(),
        "c": types.Float(),
    }


def test_save_load_pydantic():
    class TestPydanticClass(pydantic.BaseModel):
        a: int

    obj = TestPydanticClass(a=1)
    n = weave.save(obj)
    recovered = weave.use(n)
    assert recovered.a == 1


def test_pydantic_saveload(client):
    class Object(pydantic.BaseModel):
        name: Optional[str] = "hello"
        description: Optional[str] = None

    class A(Object):
        model_name: str

    class B(A):
        pass

    a = B(name="my-a", model_name="my-model")

    a_type = weave.type_of(a)
    assert a_type.root_type_class().__name__ == "A"

    weave.publish(a, name="my-a")

    a2 = weave.ref("my-a").get()
    assert a2.name == "my-a"
    assert a2.description is None
    assert a2.model_name == "my-model"


def test_pydantic_nested_type():
    class Child(pydantic.BaseModel):
        a: str

    class Parent(pydantic.BaseModel):
        child: Child
        b: int

    p = Parent(child=Child(a="hello"), b=1)
    t = weave.type_of(p)

    assert str(t) == "Parent(child=Child(a=String()), b=Int())"
    assert t.to_dict() == {
        "type": "Parent",
        "_base_type": {"type": "Object"},
        "_relocatable": True,
        "_is_object": True,
        "child": {
            "type": "Child",
            "_base_type": {"type": "Object"},
            "_relocatable": True,
            "_is_object": True,
            "a": "string",
        },
        "b": "int",
    }


def test_pydantic_v1(client):
    from pydantic import v1

    class MyV1Object(v1.BaseModel):
        val: int

    class MyWeaveObject(weave.Object):
        inner: Any
        # inner: MyV1Object # v1 properties not yet supported

    @weave.op
    def get_inner(obj: MyWeaveObject) -> int:
        return obj.inner.val

    res = get_inner(MyWeaveObject(inner=MyV1Object(val=1)))
    assert res == 1



import contextlib
import os
from unittest.mock import Mock

import openai
import pytest
import pytest_asyncio
from openai import AsyncStream, Stream
from openai.types.chat.chat_completion import ChatCompletion, Choice
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk, ChoiceDelta
from openai.types.chat.chat_completion_chunk import Choice as ChunkChoice
from openai.types.completion_usage import CompletionUsage

import weave
from weave.legacy.weave.monitoring import init_monitor
from weave.legacy.weave.monitoring.openai import util
from weave.legacy.weave.monitoring.openai.models import *
from weave.legacy.weave.monitoring.openai.models import Context
from weave.legacy.weave.monitoring.openai.openai import patch, unpatch
from weave.legacy.weave.monitoring.openai.util import Context
from weave.legacy.weave.wandb_interface.wandb_stream_table import StreamTable


@pytest.fixture
def chat_completion_request_message():
    return [
        ChatCompletionRequestMessage(
            content="Tell me a joke", role="system", function_call=None, tool_calls=None
        )
    ]


@pytest.fixture
def streaming_chat_completion_messages():
    return [
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content="",
                        function_call=None,
                        role="assistant",
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content="Why", function_call=None, role=None, tool_calls=None
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content=" don", function_call=None, role=None, tool_calls=None
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content="'t", function_call=None, role=None, tool_calls=None
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content=" scientists",
                        function_call=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content=" trust", function_call=None, role=None, tool_calls=None
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content=" atoms", function_call=None, role=None, tool_calls=None
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content="?\n\n", function_call=None, role=None, tool_calls=None
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content="Because",
                        function_call=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content=" they", function_call=None, role=None, tool_calls=None
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content=" make", function_call=None, role=None, tool_calls=None
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content=" up", function_call=None, role=None, tool_calls=None
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content=" everything",
                        function_call=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content="!", function_call=None, role=None, tool_calls=None
                    ),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
        ChatCompletionChunk(
            id="chatcmpl-blank",
            choices=[
                ChunkChoice(
                    delta=ChoiceDelta(
                        content=None, function_call=None, role=None, tool_calls=None
                    ),
                    finish_reason="stop",
                    index=0,
                )
            ],
            created=1700686161,
            model="gpt-3.5-turbo-0613",
            object="chat.completion.chunk",
        ),
    ]


class MockAsyncResponse:
    """This emulates an SSE response"""

    def __init__(self, chunks: List):
        self._chunks = iter(chunks)

    async def aiter_lines(self):
        i = 0
        for chunk in self._chunks:
            i += 1
            yield f"data: {chunk.model_dump_json()}\n"
            yield "\n"
        yield "data: [DONE]\n"
        yield "\n"

    async def aiter_bytes(self):
        async for line in self.aiter_lines():
            yield line.encode("utf-8")


class MockAsyncStream(AsyncStream):
    def __init__(self, chunks: List):
        self._chunks = iter(chunks)

        def process_response_data(*, data: object, **kwargs):
            return ChatCompletionChunk.model_validate(data)

        def make_sse_decoder():
            from openai._streaming import SSEDecoder

            return SSEDecoder()

        super().__init__(
            cast_to=ChatCompletionChunk,
            client=Mock(
                _process_response_data=process_response_data,
                _make_sse_decoder=make_sse_decoder,
            ),
            response=MockAsyncResponse(chunks),
        )


@pytest.fixture
def async_streaming_chat_completion_messages(streaming_chat_completion_messages):
    return MockAsyncStream(streaming_chat_completion_messages)


@pytest.fixture
def reassembled_chat_completion_message():
    return ChatCompletion(
        id="chatcmpl-blank",
        choices=[
            Choice(
                # logprobs included here because latest versions of pydantic
                # require optional fields without default values to be included
                # in the constructor
                logprobs=None,
                finish_reason="stop",
                index=0,
                message=ChatCompletionMessage(
                    content="Why don't scientists trust atoms?\n\nBecause they make up everything!",
                    role="assistant",
                    tool_calls=None,
                ),
            )
        ],
        created=1700686161,
        model="gpt-3.5-turbo-0613",
        object="chat.completion",
        usage=CompletionUsage(completion_tokens=19, prompt_tokens=10, total_tokens=29),
    )


@pytest.fixture
def openai_client():
    c = Mock(spec=openai.OpenAI)
    c.chat = Mock()
    c.chat.completions = Mock()
    c.chat.completions.create = Mock()
    return c


@pytest.fixture
def streaming_client(openai_client, streaming_chat_completion_messages):
    openai_client.chat.completions.create.return_value = iter(
        streaming_chat_completion_messages
    )
    return openai_client


@pytest.fixture
def teardown():
    yield
    unpatch()


class MockSyncResponse:
    """This emulates a synchronous SSE response"""

    def __init__(self, chunks: List):
        self._chunks = iter(chunks)

    def iter_lines(self):
        i = 0
        for chunk in self._chunks:
            i += 1
            yield f"data: {chunk.model_dump_json()}\n"
            yield "\n"
        yield "data: [DONE]\n"
        yield "\n"

    def iter_bytes(self):
        for line in self.iter_lines():
            yield line.encode("utf-8")


class MockStream(Stream):
    def __init__(self, chunks: List):
        self._chunks = iter(chunks)

        def process_response_data(*, data: object, **kwargs):
            return ChatCompletionChunk.model_validate(data)

        def make_sse_decoder():
            from openai._streaming import SSEDecoder

            return SSEDecoder()

        super().__init__(
            cast_to=ChatCompletionChunk,
            client=Mock(
                _process_response_data=process_response_data,
                _make_sse_decoder=make_sse_decoder,
            ),
            response=MockSyncResponse(chunks),
        )


@pytest.fixture
def mocked_streaming_create(streaming_chat_completion_messages):
    # Mock the base create method
    return Mock(return_value=MockStream(streaming_chat_completion_messages))


@pytest_asyncio.fixture
async def mocked_async_streaming_create(async_streaming_chat_completion_messages):
    async def mocked_create(*args, **kwargs):
        return async_streaming_chat_completion_messages

    return mocked_create


@pytest.fixture
def mocked_create(reassembled_chat_completion_message):
    # Mock the base create method
    return Mock(return_value=reassembled_chat_completion_message)


def make_stream_table(*args, **kwargs):
    # Unit test backend does not support async logging
    return StreamTable(*args, **kwargs, _disable_async_file_stream=True)


##########


# def test_log_to_span_basic(
#     user_by_api_key_in_env,
#     mocked_create,
#     teardown,
#     reassembled_chat_completion_message,
#     client,
# ):
#     stream_name = "monitoring"
#     project = "openai"
#     entity = user_by_api_key_in_env.username

#     streamtable = make_stream_table(
#         stream_name, project_name=project, entity_name=entity
#     )
#     chat_completions = weave.legacy.weave.monitoring.openai.openai.ChatCompletions(
#         mocked_create
#     )
#     create_input = dict(
#         model="gpt-3.5-turbo",
#         messages=[{"role": "system", "content": "Tell me a joke"}],
#     )
#     result = chat_completions.create(**create_input)
#     streamtable.finish()

#     call = client.get_calls()[0]
#     inputs = {k: v for k, v in call.inputs.items() if not k.startswith("_")}
#     outputs = {k: v for k, v in call.output.items() if not k.startswith("_")}

#     inputs_expected = create_input
#     assert inputs == inputs_expected

#     outputs_expected = reassembled_chat_completion_message.dict(exclude_unset=True)
#     assert outputs == outputs_expected


# def test_log_to_span_streaming(
#     user_by_api_key_in_env,
#     mocked_streaming_create,
#     teardown,
#     reassembled_chat_completion_message,
#     client,
# ):
#     chat_completions = weave.legacy.weave.monitoring.openai.openai.ChatCompletions(
#         mocked_streaming_create
#     )
#     create_input = dict(
#         model="gpt-3.5-turbo",
#         messages=[{"role": "system", "content": "Tell me a joke"}],
#         stream=True,
#     )
#     stream = chat_completions.create(**create_input)
#     for x in stream:
#         ...

#     call = client.get_calls()[0]
#     inputs = {k: v for k, v in call.inputs.items() if not k.startswith("_")}
#     outputs = {k: v for k, v in call.output.items() if not k.startswith("_")}

#     inputs_expected = create_input
#     assert inputs == inputs_expected

#     outputs_expected = reassembled_chat_completion_message.dict(exclude_unset=True)
#     assert outputs == outputs_expected


# @pytest.mark.asyncio
# async def test_log_to_span_async_streaming(
#     user_by_api_key_in_env,
#     mocked_async_streaming_create,
#     teardown,
#     reassembled_chat_completion_message,
#     client,
# ):
#     chat_completions = weave.legacy.weave.monitoring.openai.openai.AsyncChatCompletions(
#         mocked_async_streaming_create
#     )
#     create_input = dict(
#         model="gpt-3.5-turbo",
#         messages=[{"role": "system", "content": "Tell me a joke"}],
#         stream=True,
#     )
#     stream = await chat_completions.create(**create_input)
#     async for x in stream:
#         ...

#     call = client.get_calls()[0]
#     inputs = {k: v for k, v in call.inputs.items() if not k.startswith("_")}
#     outputs = {k: v for k, v in call.output.items() if not k.startswith("_")}

#     inputs_expected = create_input
#     assert inputs == inputs_expected

#     outputs_expected = reassembled_chat_completion_message.dict(exclude_unset=True)
#     assert outputs == outputs_expected


@contextlib.contextmanager
def async_disabled():
    current = os.environ.get("WEAVE_DISABLE_ASYNC_FILE_STREAM")
    os.environ["WEAVE_DISABLE_ASYNC_FILE_STREAM"] = "true"
    try:
        yield
    finally:
        if current is None:
            del os.environ["WEAVE_DISABLE_ASYNC_FILE_STREAM"]
        else:
            os.environ["WEAVE_DISABLE_ASYNC_FILE_STREAM"] = current



import pandas as pd

from weave.legacy.weave import api as weave
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.ops_primitives import pandas_ as op_pandas


def test_save_dataframe():
    data = [["cat", 10], ["dog", 3], ["cat", 1]]
    df = pd.DataFrame(data, columns=["class", "age"])
    ref = weave.save(df, "my-df")
    df2 = weave.use(ref)
    assert df.equals(df2)


def test_save_dataframe_table():
    data = [["cat", 10], ["dog", 3], ["cat", 1]]
    df = pd.DataFrame(data, columns=["class", "age"])
    # print("DF", op_pandas.DataFrameTable)
    df_table = op_pandas.DataFrameTable(df)
    ref = weave.save(df_table, "my-df-table")
    df_table2 = weave.use(ref)
    assert df_table._df.equals(df_table2._df)


def test_dataframe_type():
    data = [["cat", 10, 0.1], ["dog", 3, 0.2], ["cat", 1, 0.3]]
    df = pd.DataFrame(data, columns=["class", "age", "loss"])
    assert types.TypeRegistry.type_of(df).object_type == types.TypedDict(
        {
            "class": types.String(),
            "age": types.Int(),
            "loss": types.Float(),
        }
    )



import bertviz

import weave
from weave.legacy.weave.ecosystem import huggingface


@weave.op()
def head_view(attention: huggingface.ModelOutputAttention) -> weave.legacy.weave.ops.Html:
    # All the information we need is attached to the ModelOutputAttention object.
    # This is important. In Weave, types should "stand alone", meaning they should
    # contain references to any information that is necessary for their use.

    # we can walk to model_output from attention
    model_output = attention.model_output

    # we can walk to model_input from model_output
    encoded_model_input = model_output._encoded_model_input

    # we can walk to the model
    model = model_output._model

    tokens = model.tokenizer().convert_ids_to_tokens(encoded_model_input[0])

    # Finally call bertviz, which gives us back html.
    html = bertviz.head_view(attention._attention, tokens, html_action="return")

    # The .data attribute contains the html string. Wrap it in the weave Html type.

    # TODO: this would read better as weave.types.Html I think.
    return weave.legacy.weave.ops.Html(html.data)


@weave.type()
class BertvizHeadView(weave.Panel):
    id = "BertvizHeadView"
    # Panels specify their input type.
    input_node: weave.Node[huggingface.ModelOutputAttention]

    @weave.op()
    def render(self) -> weave.legacy.weave.panels.PanelHtml:
        # This is a lazy call! It doesn't execute anything
        html = head_view(self.input_node)

        # We add the lazy call as input to the returned Html panel. Nothing has been
        # computed so far. The UI's Html panel will perform a useNodeValue operation on its
        # input node. Only then will the head_view function finally be called.
        return weave.legacy.weave.panels.PanelHtml(html)


@weave.op()
def model_view(attention: huggingface.ModelOutputAttention) -> weave.legacy.weave.ops.Html:
    # Parallels head_view() to visualize the full matrix of attention heads as rows
    # and layers as columns for each attention map

    # walk to model from attention
    model_output = attention.model_output
    encoded_model_input = model_output._encoded_model_input
    model = model_output._model

    tokens = model.tokenizer().convert_ids_to_tokens(encoded_model_input[0])
    html = bertviz.model_view(attention._attention, tokens, html_action="return")

    # TODO: this would read better as weave.types.Html I think.
    return weave.legacy.weave.ops.Html(html.data)


@weave.type()
class BertvizModelView(weave.Panel):
    id = "BertvizModelView"
    # Panels specify their input type.
    input_node: weave.Node[huggingface.ModelOutputAttention]

    @weave.op()
    def model_view_panel_render(self) -> weave.legacy.weave.panels.PanelHtml:
        html = model_view(self.input_node)
        return weave.legacy.weave.panels.PanelHtml(html)



import typing

import weave
import weave.legacy.weave
from weave.legacy.weave import val_const


class ScenarioResult(typing.TypedDict):
    scenario_id: str
    metric1: float
    metric2: float
    metric3: float
    metric4: float
    metric5: float
    metric6: float


class MetricsBankInput(typing.TypedDict):
    baseline: list[ScenarioResult]
    candidate: list[ScenarioResult]


@weave.type()
class MetricsBankPanel(weave.Panel):
    id = "MetricsBankPanel"
    input_node: weave.Node[MetricsBankInput]

    @weave.op()
    def render(self) -> weave.legacy.weave.panels.Each:
        input = typing.cast(MetricsBankInput, self.input_node)

        baseline = input["baseline"]
        candidate = input["candidate"]

        joined = weave.legacy.weave.ops.join_all(
            weave.legacy.weave.ops.make_list(l0=baseline, l1=candidate),
            lambda row: row["scenario_id"],
            False,
        )

        # The output type of keys includes the keys (its List["scenario_id" | "metric1" | "metric2" | "metric3"])
        joined_keys = joined[0].keys()

        # The output type of difference is List["metric1" | "metric2" | "metric3"]
        metrics = weave.legacy.weave.ops.difference(
            joined_keys, [val_const.const("scenario_id")]
        )

        # TODO: broken
        return weave.legacy.weave.panels.Each(
            metrics,
            render=lambda metric_name: weave.legacy.weave.panels.Group(
                items={
                    "title": metric_name,
                    "plot": weave.legacy.weave.panels.Plot(
                        joined,
                        # TODO: bring this back
                        # title=metric_name,
                        # The [metric_name] pick operations correctly product list[float], since
                        # we know metric_name is not scenario_id in the type system.
                        # If this produced list[float | str], PanelPlot would not know how to render
                        # the data.
                        x=lambda row: row[metric_name][0],
                        # x_title="baseline",
                        y=lambda row: row[metric_name][1],
                        # y_title="candidate",
                    ),
                },
            ),
        )  # type: ignore



import atexit
import concurrent.futures
import logging
import threading
from concurrent.futures import Future
from typing import Any, Callable, TypeVar

from weave.trace.context import get_raise_on_captured_errors

T = TypeVar("T")

MAX_WORKER_DEFAULT = 2**3  # 8 workers to not overwhelm the DB

logger = logging.getLogger(__name__)


class AsyncJobQueue:
    """A queue for managing asynchronous job execution.

    This class provides a thread-safe way to submit and execute jobs asynchronously
    using a ThreadPoolExecutor. It ensures proper shutdown of the executor when the
    object is deleted or when the program exits. If jobs are submitted after shutdown,
    the queue will restart automatically.

    Attributes:
        executor (concurrent.futures.ThreadPoolExecutor): The executor used to run jobs.
        _lock (threading.Lock): A lock to ensure thread-safe operations.
        _active_jobs (set): A set to keep track of active jobs and callbacks.
        _max_workers (int): The maximum number of worker threads to use.

    Args:
        max_workers (int): The maximum number of worker threads to use. Defaults to 8.
    """

    def __init__(self, max_workers: int = MAX_WORKER_DEFAULT) -> None:
        self._max_workers = max_workers
        self._lock = threading.Lock()
        self._active_jobs: set[Future] = set()
        self.executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=self._max_workers, thread_name_prefix="AsyncJobQueue"
        )
        atexit.register(self._at_exit_handler)

    def submit_job(
        self, func: Callable[..., T], *args: Any, **kwargs: Any
    ) -> Future[T]:
        """Submits a job to be executed asynchronously.

        If the queue has been shut down, it will be restarted automatically.

        Args:
            func: The function to be executed.
            *args: Positional arguments to pass to the function.
            **kwargs: Keyword arguments to pass to the function.

        Returns:
            A Future representing the execution of the job.

        Example usage:

        ```python
        def example_job(x):
            time.sleep(1)  # Simulate some work
            return x * 2

        queue = AsyncJobQueue(max_workers=4)

        # Submit multiple jobs
        futures = [queue.submit_job(example_job, i) for i in range(5)]

        # Wait for all jobs to complete and get results
        results = [future.result() for future in futures]
        print(results)  # Output: [0, 2, 4, 6, 8]

        # Shutdown the queue when done
        queue.shutdown()
        ```
        """
        with self._lock:
            future = self.executor.submit(func, *args, **kwargs)
            self._active_jobs.add(future)

        def callback(f: Future[T]) -> None:
            with self._lock:
                self._active_jobs.remove(f)
            exception = f.exception()
            if exception:
                logger.error(f"Job failed with exception: {exception}")
                if get_raise_on_captured_errors():
                    raise

        future.add_done_callback(callback)
        return future

    def _at_exit_handler(self) -> None:
        """Ensures the executor is shut down when the program exits."""
        try:
            self.executor.shutdown(wait=True)
        except Exception as e:
            logger.error(f"Error shutting down executor: {e}")
            if get_raise_on_captured_errors():
                raise

    def flush(self) -> None:
        """Waits for all currently submitted jobs to complete.

        This method blocks until all active jobs in the queue at the time of calling
        have finished executing. It prevents new jobs from interfering with the flush operation.
        """
        active_jobs = []
        with self._lock:
            active_jobs = list(self._active_jobs)

        for future in concurrent.futures.as_completed(active_jobs):
            try:
                future.result()
            except Exception as e:
                logger.error(f"Job failed during flush: {e}")
                if get_raise_on_captured_errors():
                    raise



"""Pluggable object serializers for Weave

Example:

```
import faiss


from weave.trace import serializer


def save_instance(obj: faiss.Index, artifact, name: str) -> None:
    with artifact.writeable_file_path(f"{name}.faissindex") as write_path:
        faiss.write_index(obj, write_path)


def load_instance(
    artifact,
    name: str,
) -> faiss.Index:
    return faiss.read_index(artifact.path(f"{name}.faissindex"))


serializer.register_serializer(faiss.Index, save_instance, load_instance)
```

After the register_serializer call, if the user tries to save an object
that has a faiss.Index attribute, Weave will store the full FAISS index
instead of just a repr string.

We will also save the load_instance method as an op, and add a reference
to the load op from the saved object, so that the object can be correctly
deserialized in a Python runtime that does not have the serializer
registered.
"""

from dataclasses import dataclass
from typing import Any, Callable, Optional


@dataclass
class Serializer:
    target_class: type
    save: Callable
    load: Callable

    # Added to provide a function to check if an object is an instance of the
    # target class because protocol isinstance checks can fail in python3.12+
    instance_check: Optional[Callable[[Any], bool]] = None

    def id(self) -> str:
        ser_id = self.target_class.__module__ + "." + self.target_class.__name__
        if ser_id.startswith("weave."):
            # Special case for weave.Op (which is current weave.trace.op.Op).
            # The id is just Op, since we've already already stored this as
            # "Op" in the database.
            if ser_id.endswith(".Op"):
                return "Op"
        return ser_id


SERIALIZERS = []


def register_serializer(
    target_class: type,
    save: Callable,
    load: Callable,
    instance_check: Optional[Callable[[Any], bool]] = None,
) -> None:
    SERIALIZERS.append(Serializer(target_class, save, load, instance_check))


def get_serializer_by_id(id: str) -> Optional[Serializer]:
    for serializer in SERIALIZERS:
        if serializer.id() == id:
            return serializer
    return None


def get_serializer_for_obj(obj: Any) -> Optional[Serializer]:
    for serializer in SERIALIZERS:
        if serializer.instance_check and serializer.instance_check(obj):
            return serializer
        elif isinstance(obj, serializer.target_class):
            return serializer
    return None



import asyncio

import pytest

from weave.legacy.weave import filesystem, io_service


@pytest.mark.timeout(10)
@pytest.mark.asyncio
@pytest.mark.parametrize("process", [True, False])
async def test_io_service_async_client(io_server_factory, process):
    server: io_service.Server = io_server_factory(process)
    client = io_service.AsyncClient(
        server=server,
    )

    loop = asyncio.get_running_loop()

    results = []
    tasks = set()
    async with client.connect() as conn:
        for _ in range(10):
            task = loop.create_task(conn.sleep(0.1))
            tasks.add(task)

            def task_done_callback(fut):
                results.append(fut.result())

            task.add_done_callback(task_done_callback)
        await asyncio.wait(tasks)

    assert len(results) == 10

    for result in results:
        assert result == 0.1

    assert not conn.connected
    assert len(server.client_response_queues) == 0


@pytest.mark.timeout(10)
@pytest.mark.asyncio
@pytest.mark.parametrize("process", [True, False])
async def test_io_service_sync_client(io_server_factory, process):
    server: io_service.Server = io_server_factory(process)
    fs = filesystem.get_filesystem()
    client = io_service.SyncClient(
        server=server,
        fs=fs,
    )

    results = []
    for _ in range(10):
        results.append(client.sleep(0.1))

    assert len(results) == 10

    # if serial, this should be greater than 1 second (10 * 0.1)
    # if executed concurrently, this should be between 0.1 and 0.2 (0.1 + overhead)
    for result in results:
        assert result == 0.1

    assert len(server.client_response_queues) == 0



import os
import time

from weave.legacy.weave.wandb_interface.wandb_stream_table import StreamTable


def main(table_name, project_name, entity_name, sample_limit, gap_ms):
    st = StreamTable(
        table_name,
        project_name=project_name,
        entity_name=entity_name,
    )

    for i in range(sample_limit):
        time.sleep(gap_ms / 1000)
        st.log({"val": i})
        print({"val": i})

    st.finish()


if __name__ == "__main__":
    main(
        os.environ.get("TABLE_NAME", "stream"),
        os.environ.get("PROJECT_NAME", "stream_test"),
        os.environ.get("ENTITY_NAME", "timssweeney"),
        int(os.environ.get("SAMPLE_LIMIT", 1000)),
        int(os.environ.get("GAP_MS", 250)),
    )



import contextlib
import contextvars
import dataclasses
import typing

from weave.legacy.weave import server_interface

if typing.TYPE_CHECKING:
    from weave.legacy.weave import client_interface, uris
    from weave.legacy.weave.graph_client import GraphClient


# colab currently runs ipykernel < 6.0.  This resets context on every
# execution, see: https://github.com/ipython/ipykernel/pull/632.  We
# maintain a global context to work around this.
# NOTE: This logic assumes all ContextVars will exist as globals
# in this module with a leading underscore.
patch_context = False
try:
    import ipykernel
    from IPython.core.getipython import get_ipython

    if ipykernel.version_info[0] < 6:
        patch_context = True
except ImportError:
    pass

if patch_context:
    _context: typing.Dict[str, typing.Any] = dict()

    def weave_pre_run():
        glbs = globals()
        for k, v in _context.items():
            var = glbs.get("_" + k)
            if var is not None:
                var.set(v)

    def weave_post_run():
        _context.clear()
        for k, v in contextvars.copy_context().items():
            _context[k.name] = v

    ipython = get_ipython()
    if ipython is not None:
        # Incase this module is loaded multiple times
        for h in ipython.events.callbacks["pre_run_cell"]:
            if h.__name__ == "weave_pre_run":
                ipython.events.unregister("pre_run_cell", h)
        for h in ipython.events.callbacks["post_run_cell"]:
            if h.__name__ == "weave_post_run":
                ipython.events.unregister("post_run_cell", h)
        ipython.events.register("pre_run_cell", weave_pre_run)
        ipython.events.register("post_run_cell", weave_post_run)


# Set to the op uri if we're in the process of loading
# an op from an artifact.
_no_op_register: contextvars.ContextVar[typing.Optional[bool]] = contextvars.ContextVar(
    "loading_op_location", default=None
)


# Set to true if we're in the process of loading builtin functions
# this prevents us from storing the op as an artifact
_loading_built_ins: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "loading_built_ins", default=False
)


@contextlib.contextmanager
def loading_builtins(builtins):
    token = _loading_built_ins.set(builtins)
    try:
        yield _loading_built_ins.get()
    finally:
        _loading_built_ins.reset(token)


@contextlib.contextmanager
def no_op_register():
    token = _no_op_register.set(True)
    try:
        yield _no_op_register.get()
    finally:
        _no_op_register.reset(token)


def get_no_op_register():
    return _no_op_register.get()


def set_loading_built_ins(val=True) -> contextvars.Token:
    return _loading_built_ins.set(val)


def clear_loading_built_ins(token) -> None:
    _loading_built_ins.reset(token)


def get_loading_built_ins() -> bool:
    return _loading_built_ins.get()


_analytics_enabled: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "analytics_enabled", default=True
)

_weave_client: contextvars.ContextVar[
    typing.Optional["client_interface.ClientInterface"]
] = contextvars.ContextVar("weave_client", default=None)

_monitor_disabled: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "monitor_disabled", default=False
)


@contextlib.contextmanager
def client(client: "client_interface.ClientInterface"):
    client_token = _weave_client.set(client)
    try:
        yield client
    finally:
        _weave_client.reset(client_token)


def get_client() -> typing.Optional["client_interface.ClientInterface"]:
    return _weave_client.get()


def set_client(client: "client_interface.ClientInterface"):
    _weave_client.set(client)


_http_server: contextvars.ContextVar[
    typing.Optional["server_interface.BaseServer"]
] = contextvars.ContextVar("http_server", default=None)


@contextlib.contextmanager
def server(server: "server_interface.BaseServer"):
    server_token = _http_server.set(server)
    try:
        yield server
    finally:
        _http_server.reset(server_token)


def get_server() -> typing.Optional["server_interface.BaseServer"]:
    return _http_server.get()


def set_server(server: "server_interface.BaseServer"):
    _http_server.set(server)


_frontend_url: contextvars.ContextVar[typing.Optional[str]] = contextvars.ContextVar(
    "frontend_url", default=None
)


@contextlib.contextmanager
def monitor_disabled():
    token = _monitor_disabled.set(True)
    yield
    _monitor_disabled.reset(token)


def get_frontend_url() -> typing.Optional[str]:
    return _frontend_url.get()


def set_frontend_url(url: str):
    _frontend_url.set(url)


_eager_mode: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "eager_mode", default=True
)


@contextlib.contextmanager
def eager_execution():
    eager_token = _eager_mode.set(True)
    try:
        yield
    finally:
        _eager_mode.reset(eager_token)


@contextlib.contextmanager
def lazy_execution():
    eager_token = _eager_mode.set(False)
    try:
        yield
    finally:
        _eager_mode.reset(eager_token)


@contextlib.contextmanager
def set_eager_mode(eager: bool):
    eager_token = _eager_mode.set(eager)
    try:
        yield
    finally:
        _eager_mode.reset(eager_token)


def eager_mode() -> bool:
    return _eager_mode.get()


@contextlib.contextmanager
def analytics_disabled():
    analytics_token = _analytics_enabled.set(False)
    try:
        yield
    finally:
        _analytics_enabled.reset(analytics_token)


def analytics_enabled() -> bool:
    return _analytics_enabled.get()


def disable_analytics() -> contextvars.Token:
    return _analytics_enabled.set(False)


_client_cache_key: contextvars.ContextVar[
    typing.Optional[str]
] = contextvars.ContextVar("client_cache_key", default=None)


@contextlib.contextmanager
def set_client_cache_key(key: typing.Optional[str] = None):
    token = _client_cache_key.set(key)
    try:
        yield
    finally:
        _client_cache_key.reset(token)


def get_client_cache_key():
    return _client_cache_key.get()


# Context for wandb api
# Instead of putting this in a shared file, we put it directly here
# so that the patching at the top of this file will work correctly
# for this symbol.
@dataclasses.dataclass
class WandbApiContext:
    user_id: typing.Optional[str] = None
    api_key: typing.Optional[str] = None
    headers: typing.Optional[dict[str, str]] = None
    cookies: typing.Optional[dict[str, str]] = None

    @classmethod
    def from_json(cls, json: typing.Any) -> "WandbApiContext":
        return cls(**json)

    def to_json(self) -> typing.Any:
        return dataclasses.asdict(self)


## wandb_api.py context
_wandb_api_context: contextvars.ContextVar[
    typing.Optional[WandbApiContext]
] = contextvars.ContextVar("wandb_api_context", default=None)

## urls.py Context
_use_local_urls: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "use_local_urls", default=False
)

## graph_client_context.py Context
_graph_client: contextvars.ContextVar[
    typing.Optional["GraphClient"]
] = contextvars.ContextVar("graph_client", default=None)


_cache_prefix_context: contextvars.ContextVar[
    typing.Optional[str]
] = contextvars.ContextVar("cache_prefix", default=None)

_ref_tracking_enabled: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "ref_tracking_enabled", default=False
)


def ref_tracking_enabled() -> bool:
    return _ref_tracking_enabled.get()


@contextlib.contextmanager
def ref_tracking(enabled: bool):
    token = _ref_tracking_enabled.set(enabled)
    yield _ref_tracking_enabled.get()
    _ref_tracking_enabled.reset(token)


_serverless_io_service: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "_serverless_io_service", default=False
)


def serverless_io_service() -> bool:
    return _serverless_io_service.get()


# Throw an error if op saving encounters an unknonwn condition.
# The default behavior is to warn.
_legacy_strict_op_saving: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "_legacy_strict_op_saving", default=False
)


def get_legacy_strict_op_saving() -> bool:
    return _legacy_strict_op_saving.get()


@contextlib.contextmanager
def legacy_strict_op_saving(enabled: bool):
    token = _legacy_strict_op_saving.set(enabled)
    yield _legacy_strict_op_saving.get()
    _legacy_strict_op_saving.reset(token)



"""Contains the version of the Weave SDK.

This version is used by pyproject.toml to specify the version of the Weave SDK and
imported by the library for runtime access. It must be kept in sync with pyproject.toml,
specifically the [tool.bumpversion] section:

```
[tool.bumpversion]
current_version = "M.m.p-dev0"
```

We use Semantic Versioning (https://semver.org/). The version number is in the format
MAJOR.MINOR.PATCH[-PRERELEASE].

- MAJOR version when you make incompatible API changes,
- MINOR version when you add functionality in a backwards compatible manner, and
- PATCH version when you make backwards compatible bug fixes.

For the most part, we are incrementing PATCH until we complete the core functionality.

As specified by Semantic Versioning, the PRERELEASE version is optional and can be
appended to the PATCH version. Specifically:

```
When major, minor, and patch are equal, a pre-release version has lower precedence than a normal version:

Example: 1.0.0-alpha < 1.0.0.
```

The intention is to have a PRERELEASE version of the form of `dev0` for nearly every commit
on the main branch. However, the released commit will not have a PRERELEASE version. For example:

* Development happens on X.Y.Z-dev0
* Release commit bumps the version to >=X.Y.Z, noted X'.Y'.Z', tagging such commit
* Development continues on (>X.Y.Z)-dev0.
    * Note: if X' == X, Y' == Y, and Z' == Z, then the version is bumped to X'.Y'.(Z'+1)-dev0. Else, the version is bumped to X'.Y'.Z'-dev0.


This is all facilitated using `make bumpversion` which is a wrapper around `bump-my-version`.

Please see the `bump-my-version` documentation for more information:
https://github.com/callowayproject/bump-my-version. Additional configuration can be found in
`pyproject.toml`.

"""

VERSION = "0.51.9"



# Functions for optimizing table access sub-DAGs.

import typing

from weave.legacy.weave import errors
from weave.legacy.weave import stitch

KeyTree = typing.Dict[str, "KeyTree"]  # type:ignore


# Tree merges b into a IN PLACE
def tree_merge(a: KeyTree, b: KeyTree) -> None:
    for k, v in b.items():
        a_val = a.setdefault(k, {})
        if isinstance(v, dict):
            tree_merge(a_val, v)
        else:
            a_val[k] = v


def get_projection(obj: stitch.ObjectRecorder) -> KeyTree:
    """Given an object returned by stitch, return a tree of all accessed columns."""
    from weave.legacy.weave import _dict_utils

    cols: KeyTree = {}
    all_keys = False
    for call in obj.calls:
        if call.node.from_op.name.endswith("pick"):
            key = call.inputs[1].val
            if key is None:
                raise errors.WeaveInternalError("non-const not yet supported")
            key_path = _dict_utils.split_escaped_string(key)
            item_root = cols
            for k in key_path:
                if k != "*":
                    item_root = item_root.setdefault(_dict_utils.escape_dots(k), {})
            tree_merge(item_root, get_projection(call.output))
        elif call.node.from_op.name.endswith("__getattr__"):
            key = call.inputs[1].val
            if key is None:
                raise errors.WeaveInternalError("non-const not yet supported")
            tree_merge(cols.setdefault(key, {}), get_projection(call.output))
        elif call.node.from_op.name.endswith("keytypes"):
            all_keys = True
        else:
            tree_merge(cols, get_projection(call.output))
    if all_keys:
        cols = {}
    return cols



import collections
import contextlib
import contextvars
import copy
import inspect
import typing
from typing import Iterator, Sequence

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import (
    weave_internal,
    errors,
    context_state,
    engine_trace,
    graph,
    language_autocall,
    memo,
    object_context,
    op_args,
    op_def_type,
    op_execute,
    pyfunc_type_util,
    uris,
)
from weave.legacy.weave.language_features.tagging import (
    opdef_util,
    process_opdef_output_type,
    process_opdef_resolve_fn,
    tagged_value_type,
)
from weave.legacy.weave.run import Run
from weave.legacy.weave.weavejs_fixes import fixup_node

if typing.TYPE_CHECKING:
    from weave.trace import weave_client
    from weave.legacy.weave.run_streamtable_span import RunStreamTableSpan


_no_refine: contextvars.ContextVar[bool] = contextvars.ContextVar(
    "_no_refine", default=False
)


def refine_enabled() -> bool:
    return not _no_refine.get()


@contextlib.contextmanager
def no_refine():
    token = _no_refine.set(True)
    try:
        yield
    finally:
        _no_refine.reset(token)


def common_name(name: str) -> str:
    return name.split("-")[-1]


# TODO: Move to weave_types after moving TaggedValueType into weave_types
def map_type(
    t: types.Type, map_fn: typing.Callable[[types.Type], types.Type]
) -> types.Type:
    if isinstance(t, types.NoneType):
        pass
    elif isinstance(t, types.Const):
        t = types.Const(map_type(t.val_type, map_fn), t.val)
    elif isinstance(t, types.UnionType):
        t = types.union(*(map_type(m, map_fn) for m in t.members))
    elif isinstance(t, types.TypedDict):
        t = types.TypedDict(
            {k: map_type(v, map_fn) for k, v in t.property_types.items()}
        )
    elif isinstance(t, tagged_value_type.TaggedValueType):
        mapped_tag = map_type(t.tag, map_fn)
        if not isinstance(mapped_tag, types.TypedDict):
            raise errors.WeaveTypeError(
                f"TaggedValueType tag must be a TypedDict, got {mapped_tag}"
            )
        t = tagged_value_type.TaggedValueType(mapped_tag, map_type(t.value, map_fn))
    elif hasattr(t, "object_type"):
        t = t.__class__(object_type=map_type(t.object_type, map_fn))  # type: ignore
    mapped_t = map_fn(t)
    if mapped_t is None:
        return t
    return mapped_t


@memo.memo
def normalize_type(t: types.Type) -> types.Type:
    # This produces equivalent types, but normalized in the way Weave0 expects.
    def _norm(t):
        # Distribute TaggedValueType over UnionType
        if isinstance(t, tagged_value_type.TaggedValueType) and isinstance(
            t.value, types.UnionType
        ):
            return types.union(
                *[
                    tagged_value_type.TaggedValueType(t.tag, m)
                    for m in t.value.members
                    if not isinstance(m, types.NoneType)
                ]
            )

    return map_type(t, _norm)


def _full_output_type(
    params: dict[str, types.Type],
    op_input0_type: types.Type,
    op_output_type: typing.Union[
        types.Type, typing.Callable[[typing.Dict[str, types.Type]], types.Type]
    ],
    param0_is_const: bool,
    param0_const_val: typing.Any,
) -> types.Type:
    # TODO: I think this would be a lot more clear using a stack, since its tail recursive
    # and we have to pass down const information.

    param0_name = list(params.keys())[0]
    param0_type = params[param0_name]
    # Note isinstance check, won't work if op for some reason handles a union of TaggedValueTypes.
    # But the only ops that handle TaggedValueTypes should be tag getters...
    # TODO: formalize?
    if isinstance(param0_type, types.Const):
        # We remove the Const type, since we're computing the output type which can't be const
        # since its derived.
        return _full_output_type(
            {**params, param0_name: param0_type.val_type},
            op_input0_type,
            op_output_type,
            param0_is_const=True,
            param0_const_val=param0_type.val,
        )
    if not isinstance(op_input0_type, types.Function) and isinstance(
        param0_type, types.Function
    ):
        return _full_output_type(
            {**params, param0_name: param0_type.output_type},
            op_input0_type,
            op_output_type,
            param0_is_const=param0_is_const,
            param0_const_val=param0_const_val,
        )
    elif not isinstance(
        op_input0_type, tagged_value_type.TaggedValueType
    ) and isinstance(param0_type, tagged_value_type.TaggedValueType):
        return tagged_value_type.TaggedValueType(
            param0_type.tag,
            _full_output_type(
                {**params, param0_name: param0_type.value},
                op_input0_type,
                op_output_type,
                param0_is_const=param0_is_const,
                param0_const_val=param0_const_val,
            ),
        )
    elif isinstance(param0_type, types.UnionType):
        return types.union(
            *(
                _full_output_type(
                    {**params, param0_name: m},
                    op_input0_type,
                    op_output_type,
                    param0_is_const=param0_is_const,
                    param0_const_val=param0_const_val,
                )
                for m in param0_type.members
            )
        )
    elif not op_input0_type.assign_type(types.NoneType()) and isinstance(
        param0_type, types.NoneType
    ):
        return types.NoneType()

    if callable(op_output_type):
        if param0_is_const:
            return op_output_type(
                {**params, param0_name: types.Const(param0_type, param0_const_val)}
            )
        return op_output_type(params)
    return op_output_type


def full_output_type(
    params: dict[str, types.Type],
    op_input0_type: types.Type,
    op_output_type: typing.Union[
        types.Type, typing.Callable[[typing.Dict[str, types.Type]], types.Type]
    ],
) -> types.Type:
    return _full_output_type(params, op_input0_type, op_output_type, False, None)


class OpDef:
    """An Op Definition.

    Must be immediately passed to Register.register_op() after construction.
    """

    input_type: op_args.OpArgs
    raw_output_type: typing.Union[
        types.Type,
        typing.Callable[[typing.Dict[str, types.Type]], types.Type],
    ]
    refine_output_type: typing.Optional["OpDef"]
    setter: typing.Optional[typing.Callable] = None
    version: typing.Optional[str]
    location: typing.Optional[uris.WeaveURI]
    is_builtin: bool = False
    weave_fn: typing.Optional[graph.Node]
    instance: typing.Union[None, graph.Node]
    hidden: bool
    pure: bool
    mutation: bool
    raw_resolve_fn: typing.Callable
    _output_type: typing.Optional[
        typing.Union[
            types.Type,
            typing.Callable[[typing.Dict[str, types.Type]], types.Type],
        ]
    ]
    plugins: typing.Optional[typing.Dict[str, typing.Any]]

    # This is required to be able to determine which ops were derived from this
    # op. Particularly in cases where we need to rename or lookup when
    # versioned, we cannot rely just on naming structure alone.
    derived_ops: typing.Dict[str, "OpDef"]
    derived_from: typing.Optional["OpDef"]

    # Only used by Tag Getter Ops
    # TODO: How should we support additional info
    # in OpDefs? Maybe we should convert to using plugins
    # like th GQL ops?
    _gets_tag_by_name: typing.Optional[str] = None

    def __init__(
        self,
        name: str,
        input_type: op_args.OpArgs,
        output_type: typing.Union[
            types.Type,
            typing.Callable[[typing.Dict[str, types.Type]], types.Type],
        ],
        resolve_fn: typing.Callable,
        refine_output_type: typing.Optional["OpDef"] = None,
        setter=None,
        hidden=False,
        render_info=None,
        pure=True,
        mutation=False,
        is_builtin: typing.Optional[bool] = None,
        weave_fn: typing.Optional[graph.Node] = None,
        *,
        plugins=None,
    ):
        self.name = name
        self.input_type = input_type
        self.raw_output_type = output_type
        self.refine_output_type = refine_output_type
        self.raw_resolve_fn = resolve_fn  # type: ignore
        self.setter = setter
        self.render_info = render_info
        self.hidden = hidden
        self.pure = pure
        self.mutation = mutation
        self.is_builtin = (
            is_builtin
            if is_builtin is not None
            else context_state.get_loading_built_ins()
        )
        self.version = None
        self.location = None
        self.instance = None
        self.derived_ops = {}
        self.derived_from = None
        self.weave_fn = weave_fn
        self._output_type = None
        self.plugins = plugins

    def __get__(self, instance, owner):
        return BoundOpDef(instance, self)

    def __call__(_self, *args, **kwargs):
        # Uses _self instead of self, because self is a typical op argument!
        if _self.mutation:
            # This is hacky. We call the resolver directly, and don't
            # convert arguments to Const nodes. There is no type checking.
            # May need to fix this, but the patterns in test_mutation2 work
            # now.
            from weave.legacy.weave import object_context

            with object_context.object_context():
                return _self.resolve_fn(*args, **kwargs)
        if context_state.eager_mode():
            return _self.eager_call(*args, **kwargs)
        else:
            return _self.lazy_call(*args, **kwargs)

    def unrefined_output_type_for_params(self, params: typing.Dict[str, graph.Node]):
        if not callable(self.output_type):
            return self.output_type
        new_input_type: dict[str, types.Type] = {}
        for k, n in params.items():
            if isinstance(n, graph.ConstNode) and not isinstance(n.type, types.Const):
                new_input_type[k] = types.Const(n.type, n.val)
            else:
                new_input_type[k] = n.type
        new_input_type = language_autocall.update_input_types(
            self.input_type, new_input_type
        )
        return self.output_type(new_input_type)

    def is_gql_root_resolver(self):
        from weave.legacy.weave import gql_op_plugin

        return self in gql_op_plugin._ROOT_RESOLVERS

    def lazy_call(_self, *args, **kwargs):
        bound_params = _self.bind_params(args, kwargs)
        # Don't try to refine if there are variable nodes, we are building a
        # function in that case!
        final_output_type: types.Type

        # If there are variables in the graph, prior to refining, try to replace them
        # with their values. This is currently only used in Panel construction paths, to
        # allow refinement to happen. We may want to move to a more general
        # Stack/Context for variables like we have in JS, but this works for now.
        def _replace_var_with_val(n):
            if isinstance(n, graph.VarNode) and hasattr(n, "_var_val"):
                return n._var_val
            return None

        refine_params = {k: graph.resolve_vars(n) for k, n in bound_params.items()}

        # Turn this on to debug scenarios where we don't correctly refine during
        # panel construction (will throw an error instead of letting it happen).
        # if refine_enabled():
        #     for arg_name, arg_node in refine_params.items():
        #         graph_vars = graph.expr_vars(arg_node)
        #         if graph_vars:
        #             raise ValueError(
        #                 "arg contained graph vars (%s): %s"
        #                 % (arg_name, [v.name for v in graph_vars])
        #             )

        if (
            refine_enabled()
            and _self.refine_output_type
            and not any(
                graph.expr_vars(arg_node) for arg_node in refine_params.values()
            )
        ):
            called_refine_output_type = _self.refine_output_type(**refine_params)
            tracer = engine_trace.tracer()  # type: ignore
            with tracer.trace("refine.%s" % _self.uri):
                # api's use auto-creates client. TODO: Fix inline import
                from weave.legacy.weave import api as api

                final_output_type = api.use(called_refine_output_type)  # type: ignore
            if final_output_type == None:
                # This can happen due to nullability. In that case, accept the output type is null
                final_output_type = types.NoneType()
            # Have to deref if in case a ref came back...
            from weave.legacy.weave import storage

            final_output_type = storage.deref(final_output_type)

            final_output_type = (
                process_opdef_output_type.process_opdef_refined_output_type(
                    final_output_type, bound_params, _self
                )
            )
        else:
            final_output_type = _self.unrefined_output_type_for_params(bound_params)
        from weave.legacy.weave import dispatch

        return dispatch.RuntimeOutputNode(final_output_type, _self.uri, bound_params)

    def eager_call(_self, *args, **kwargs):
        if (
            _self.name == "get"
            or _self.name == "root-project"
            or any(isinstance(n, graph.Node) for n in args)
        ) or (any(isinstance(n, graph.Node) for n in kwargs.values())):
            output_node = _self.lazy_call(*args, **kwargs)
            return output_node

        sig = pyfunc_type_util.get_signature(_self.raw_resolve_fn)
        params = sig.bind(*args, **kwargs)
        with object_context.object_context():
            return op_execute.execute_op(_self, params.arguments)

    def resolve_fn(__self, *args, **kwargs):
        return process_opdef_resolve_fn.process_opdef_resolve_fn(
            __self, __self.raw_resolve_fn, args, kwargs
        )

    @property
    def output_type(
        self,
    ) -> typing.Union[
        types.Type,
        typing.Callable[[typing.Dict[str, types.Type]], types.Type],
    ]:
        if self.name in [
            "save_to_ref",
            "op_get_tag_type",
            "op_make_type_tagged",
            "op_make_type_key_tag",
        ]:
            return self.raw_output_type
        else:
            if isinstance(self.input_type, op_args.OpVarArgs):
                return self.raw_output_type
            elif not isinstance(self.input_type, op_args.OpNamedArgs):
                raise errors.WeaveInternalError('unknown input type for op "%s"')
            if not self.input_type.arg_types:
                return self.raw_output_type
            first_input_type = list(self.input_type.arg_types.values())[0]

            def handle(input_type):
                param0_name = list(input_type.keys())[0]
                param0_type = input_type[param0_name]
                input_type = {**input_type, param0_name: normalize_type(param0_type)}

                def auto_tagging_output_type(input_type):
                    if isinstance(self.raw_output_type, types.Type):
                        result_type = self.raw_output_type
                    else:
                        result_type = self.raw_output_type(input_type)
                    if opdef_util.should_tag_op_def_outputs(self):
                        return tagged_value_type.TaggedValueType(
                            types.TypedDict({param0_name: param0_type}), result_type
                        )
                    return result_type

                final = full_output_type(
                    input_type, first_input_type, auto_tagging_output_type
                )

                return final

            return handle

    @property
    def concrete_output_type(self) -> types.Type:
        if callable(self.raw_output_type):
            try:
                return self.raw_output_type(self.input_type.initial_arg_types)
            except AttributeError:
                return types.UnknownType()
        return self.raw_output_type

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        if name.count("-") > 1:
            raise errors.WeaveDefinitionError(
                "Op names must have at most one hyphen. Got: %s" % name
            )
        self._name = name

    @property
    def common_name(self):
        return common_name(self.name)

    @property
    def uri(self) -> str:
        return str(self.location) if self.location is not None else self.name

    @property
    def simple_name(self):
        return uris.WeaveURI.parse(self.name).name

    @property
    def is_async(self):
        ot = self.concrete_output_type
        return (
            not callable(self.raw_output_type)
            and isinstance(ot, types.Function)
            and isinstance(ot.output_type, types.RunType)
        )

    def to_dict(self):
        output_type = self.raw_output_type
        if callable(output_type):
            # We can't callable output types to JS yet.
            if self.refine_output_type is not None:
                # if we have a callable output type and refine, js can correctly
                # make use of this op. We set our output type to concrete_output_type
                # and rely on refinement to get the correct output type.
                output_type = self.concrete_output_type.to_dict()
            else:
                # no refine and we have a callable output type. This
                # can't be used by js at the moment.
                raise errors.WeaveSerializeError(
                    "op not serializable for weavejs: %s" % self.name
                )

            # TODO: Consider removing the ability for an output_type to
            # be a function - they all must be Types or ConstNodes. Probably
            # this can be done after all the existing ops can safely be converted.
            # Once that change happens, we can do this conversion in the constructor.
            # output_type = callable_output_type_to_dict(
            #     self.input_type, output_type, self.uri
            # )
        else:
            output_type = output_type.to_dict()

        # Make callable input_type args into types.Any() for now.
        input_type = self.input_type
        if isinstance(input_type, op_args.OpVarArgs):
            # This is what we do on the frontend.
            input_types = {"manyX": "invalid"}
        else:
            arg_types = copy.copy(input_type.arg_types)
            for arg_name, arg_type in arg_types.items():
                if callable(arg_type):
                    arg_types[arg_name] = types.Any()
            input_types = op_args.OpNamedArgs(arg_types).to_dict()

        serialized = {
            # To_dict() is used to send the op list to WeaveJS.
            # We should send the full URI, not just the name, but WeaveJS
            # doesn't handle that yet, so for now, send the name.
            "name": graph.op_full_name(self),
            "input_types": input_types,
            "output_type": output_type,
        }
        if self.hidden:
            serialized["hidden"] = True
        if self.render_info is not None:
            serialized["render_info"] = self.render_info
        if self.refine_output_type is not None:
            serialized["refine_output_type_op_name"] = self.refine_output_type.name
        if self.derived_ops and "mapped" in self.derived_ops:
            serialized["mappable"] = True

        return serialized

    def __repr__(self):
        return "<OpDef: %s %s>" % (self.name, self.version)

    def bind_params(
        self, args: list[typing.Any], kwargs: dict[str, typing.Any]
    ) -> collections.OrderedDict[str, graph.Node]:
        sig = pyfunc_type_util.get_signature(self.raw_resolve_fn)
        bound_params = sig.bind(*args, **kwargs)
        bound_params_with_constants = collections.OrderedDict()
        for k, v in bound_params.arguments.items():
            param = sig.parameters[k]
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                for i, sub_v in enumerate(v):
                    if not isinstance(sub_v, graph.Node):
                        sub_v = weave_internal.const(sub_v)
                    bound_params_with_constants[str(i)] = sub_v
            elif param.kind == inspect.Parameter.VAR_KEYWORD:
                for sub_k, sub_v in v.items():
                    if not isinstance(sub_v, graph.Node):
                        sub_v = weave_internal.const(sub_v)
                    bound_params_with_constants[sub_k] = sub_v
            else:
                if not isinstance(self.input_type, op_args.OpNamedArgs):
                    raise errors.WeaveDefinitionError(
                        f"Error binding params to {self.uri} - found named params in signature, but op does not have named param args"
                    )
                param_input_type = self.input_type.arg_types[k]
                if callable(param_input_type):
                    already_bound_types = {
                        k: n.type for k, n in bound_params_with_constants.items()
                    }
                    already_bound_types = language_autocall.update_input_types(
                        self.input_type, already_bound_types
                    )
                    param_input_type = param_input_type(already_bound_types)
                if not isinstance(v, graph.Node):
                    if callable(v):
                        if not isinstance(param_input_type, types.Function):
                            raise errors.WeaveInternalError(
                                "callable passed as argument, but type is not Function. Op: %s"
                                % self.uri
                            )

                        # Allow passing in functions with fewer arguments then the op
                        # declares. E.g. for List.map I pass either of these:
                        #    lambda row, index: ...
                        #    lambda row: ...
                        inner_sig = inspect.signature(v)
                        vars = {}
                        for name in list(param_input_type.input_types.keys())[
                            : len(inner_sig.parameters)
                        ]:
                            vars[name] = param_input_type.input_types[name]

                        v = weave_internal.define_fn(vars, v)
                    else:
                        # TODO: should type-check v here.
                        v = weave_internal.make_const_node(
                            types.TypeRegistry.type_of(v), v
                        )
                bound_params_with_constants[k] = v
        return bound_params_with_constants

    def op_def_is_auto_tag_handling_arrow_op(self) -> bool:
        return isinstance(self, AutoTagHandlingArrowOpDef)


class AutoTagHandlingArrowOpDef(OpDef):
    pass


class BoundOpDef(OpDef):
    bind_self: typing.Optional[graph.Node]

    def __init__(self, bind_self: typing.Optional[graph.Node], op_def: OpDef) -> None:
        self.bind_self = bind_self
        self.op_def = op_def

    def __call__(self, *args, **kwargs):
        if self.bind_self is None:
            return self.op_def(*args, **kwargs)
        else:
            return self.op_def(self.bind_self, *args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.op_def, name)


def is_op_def(obj):
    return isinstance(obj, OpDef)


def callable_output_type_to_dict(input_type, output_type, op_name):
    if not isinstance(input_type, op_args.OpNamedArgs):
        # print(f"Failed to transform op {op_name}: Requires named args")
        return types.Any().to_dict()
    try:
        # TODO: Make this transformation more sophisticated once the type hierarchy is settled
        arg_types = {
            "input_types": types.TypedDict(
                {
                    k: types.TypeRegistry.type_of(t)
                    for k, t in input_type.arg_types.items()
                }
            )
        }
        return fixup_node(weave_internal.define_fn(arg_types, output_type)).to_json()
    except errors.WeaveMakeFunctionError as e:
        # print(f"Failed to transform op {op_name}: Invalid output type function")
        return types.Any().to_dict()


op_def_type.OpDefType.instance_classes = OpDef



# This file contains the low-level dictionary definitions for the
# data types which can be logged to StreamTables and consumed
# by various Prompts products in the Weave UI. It is rare that an
# end user would directly interact with these types. Instead, they
# are more likely to be used by someone building an integration or
# an MLOps engineer setting up a pipeline for their team.

import typing
from uuid import uuid4

from wandb.sdk.data_types.trace_tree import Span as WBSpan


# TraceSpanDict is the lowest common denominator for a span
# in the Weave system. It conforms to the OpenTelemetry spec
class TraceSpanDict(typing.TypedDict):
    # **OpenTelemetry spec**

    # The ID of the span - typically a UUID
    span_id: str

    # The name of the span - typically the name of the operation
    name: str

    # The ID of the trace this span belongs to - typically a UUID
    trace_id: str

    # The status code conforming to the OpenTelemetry spec
    # I would like to use a literal `typing.Literal["SUCCESS", "ERROR", "UNSET"]`
    # here, but then you have to type it this way everywhere and probably causes
    # more problems than it solves.
    status_code: str

    # Start and end times in seconds since the epoch
    start_time_s: float
    end_time_s: float

    # The parent span ID - typically a UUID (optional)
    # if not set, this is a root span
    parent_id: typing.Optional[str]

    # **Weave specific keys**

    # Attributes are any key value pairs associated with the span,
    # typically known before the execution operation
    attributes: typing.Optional[typing.Dict[str, typing.Any]]

    # Inputs are the parameters to the operation
    inputs: typing.Optional[typing.Dict[str, typing.Any]]

    # Output is the result of the operation
    output: typing.Optional[typing.Any]

    # Summary is a dictionary of key value pairs that summarize
    # the execution of the operation. This data is typically only
    # available after the operation has completed, as a function
    # of the output.
    summary: typing.Optional[typing.Dict[str, typing.Any]]

    # Exception is any free form string describing an exception
    # that occurred during the execution of the operation
    exception: typing.Optional[str]


# LLM Completion is a specific type of span that is used to
# represent the execution of a language model. We use a structure
# modelled after OpenAI's API to represent the inputs and outputs.


class _LLMCompletionMessage(typing.TypedDict):
    role: typing.Optional[str]
    content: typing.Optional[str]


class _LLMCompletionInputs(typing.TypedDict):
    messages: typing.Optional[typing.List[_LLMCompletionMessage]]


class _LLMCompletionChoice(typing.TypedDict):
    message: typing.Optional[_LLMCompletionMessage]
    finish_reason: typing.Optional[str]
    index: typing.Optional[int]


class _LLMCompletionOutput(typing.TypedDict):
    id: typing.Optional[str]
    object: typing.Optional[str]
    model: typing.Optional[str]
    choices: typing.Optional[typing.List[_LLMCompletionChoice]]


class _LLMCompletionSummary(typing.TypedDict):
    prompt_tokens: typing.Optional[int]
    completion_tokens: typing.Optional[int]
    total_tokens: typing.Optional[int]


# Ideally we would inherit from TraceSpanDict here, but
# mypy doesn't support that level of sophistication yet.
# The only difference between this and TraceSpanDict is
# the further specification of the inputs, output, and summary.
class LLMCompletionDict(typing.TypedDict):
    # **OpenTelemetry spec**

    # The ID of the span - typically a UUID
    span_id: str

    # The name of the span - typically the name of the operation
    name: str

    # The ID of the trace this span belongs to - typically a UUID
    trace_id: str

    # The status code conforming to the OpenTelemetry spec
    # I would like to use a literal `typing.Literal["SUCCESS", "ERROR", "UNSET"]`
    # here, but then you have to type it this way everywhere and probably causes
    # more problems than it solves.
    status_code: str

    # Start and end times in seconds since the epoch
    start_time_s: float
    end_time_s: float

    # The parent span ID - typically a UUID (optional)
    # if not set, this is a root span
    parent_id: typing.Optional[str]

    # **Weave specific keys**

    # Attributes are any key value pairs associated with the span,
    # typically known before the execution operation
    attributes: typing.Optional[typing.Dict[str, typing.Any]]

    # Inputs are the parameters to the operation
    inputs: typing.Optional[_LLMCompletionInputs]

    # Output is the result of the operation
    output: typing.Optional[_LLMCompletionOutput]

    # Summary is a dictionary of key value pairs that summarize
    # the execution of the operation. This data is typically only
    # available after the operation has completed, as a function
    # of the output.
    summary: typing.Optional[_LLMCompletionSummary]

    # Exception is any free form string describing an exception
    # that occurred during the execution of the operation
    exception: typing.Optional[str]


# We rely on all the hard work already in our W&B integration and just
# simply map the properties of the WB Span to the Weave Span
def wb_span_to_weave_spans(
    wb_span: WBSpan,
    trace_id: typing.Optional[str] = None,
    parent_id: typing.Optional[str] = None,
) -> typing.List[TraceSpanDict]:
    attributes = {**wb_span.attributes} if wb_span.attributes is not None else {}
    if hasattr(wb_span, "span_kind") and wb_span.span_kind is not None:
        attributes["span_kind"] = str(wb_span.span_kind)
    inputs = (
        wb_span.results[0].inputs
        if wb_span.results is not None and len(wb_span.results) > 0
        else None
    )
    outputs = (
        wb_span.results[0].outputs
        if wb_span.results is not None and len(wb_span.results) > 0
        else None
    )

    # Super Hack - fix merge!
    dummy_dict = {"_": ""} if parent_id == None else {}

    span_id = wb_span.span_id
    if span_id is None:
        span_id = str(uuid4())

    if (
        wb_span.start_time_ms is None
        or wb_span.end_time_ms is None
        or span_id is None
        or wb_span.name is None
    ):
        return []

    span = TraceSpanDict(
        start_time_s=wb_span.start_time_ms / 1000.0,
        end_time_s=wb_span.end_time_ms / 1000.0,
        span_id=span_id,
        name=wb_span.name,
        status_code=str(wb_span.status_code),
        trace_id=trace_id or span_id,
        parent_id=parent_id,
        # Followup: allow None in attributes & summary (there is an issue with vectorized opMerge)
        # This should be fixed before integrating inside LC
        attributes=attributes or dummy_dict,
        summary=dummy_dict,
        inputs=inputs,
        output=outputs,
        exception=wb_span.status_message
        if wb_span.status_message is not None
        else None,
    )
    spans = [span]
    for child in wb_span.child_spans or []:
        spans += wb_span_to_weave_spans(child, span["trace_id"], span["span_id"])

    return spans



import datetime
import os
import pathlib
import typing

from weave.legacy.weave.api import op
from weave import types
from weave.legacy.weave import artifact_fs, ref_base
from weave.legacy.weave.artifact_local import WORKING_DIR_PREFIX, LocalArtifact


@op(name="FilesystemArtifact-fileRefineType", hidden=True)
def artifact_file_refine_type(
    artifact: artifact_fs.FilesystemArtifact, path: str
) -> types.Type:
    return types.TypeRegistry.type_of(artifact.path_info(path))


# Warning: the return type of this is incorrect! Weave0 treats
# type 'file' (FilesystemArtifactFile) as both dir and file.
# We have a refiner to do the correct thing, but the return
# type is set to `File` so that the first non-refine compile
# path will still work.
@op(
    name="FilesystemArtifact-file",
    refine_output_type=artifact_file_refine_type,
)
def artifact_file(
    artifact: artifact_fs.FilesystemArtifact, path: str
) -> typing.Union[
    None, artifact_fs.FilesystemArtifactFile  # , artifact_fs.FilesystemArtifactDir
]:
    return artifact.path_info(path)  # type:ignore


@op(
    name="FilesystemArtifact-artifactName",
)
def artifact_artifactname(artifact: artifact_fs.FilesystemArtifact) -> str:
    return artifact.name


@op(
    name="FilesystemArtifact-artifactVersion",
)
def artifact_artifactversion(artifact: artifact_fs.FilesystemArtifact) -> str:
    return artifact.branch or artifact.version


@op(
    name="FilesystemArtifact-artifactForVersion",
)
def artifact_version_artifact_for_version(
    artifact: artifact_fs.FilesystemArtifact,
    version: str,
) -> artifact_fs.FilesystemArtifact:
    if not isinstance(artifact, LocalArtifact):
        raise ValueError(f"Artifact {artifact.name} is not a local artifact")

    if artifact._version is not None:
        if artifact._version == version:
            return artifact
        else:
            raise ValueError(f"Artifact {artifact.name} is not version {version}")

    possible_art = LocalArtifact(artifact.name, version)
    if possible_art._read_dirname == None:
        return None  # type: ignore
    return possible_art


@op(
    name="FilesystemArtifact-weaveType",
)
def artifact_version_weave_type(
    artifact: artifact_fs.FilesystemArtifact,
) -> types.Type:
    return artifact_fs.FilesystemArtifactRef(artifact, "obj").type


@op(
    name="FilesystemArtifact-metadata",
)
def filesystem_artifact_metadata(
    artifact: artifact_fs.FilesystemArtifact,
) -> dict:
    return artifact.metadata.as_dict()


@op(
    name="FilesystemArtifact-getLatestVersion",
)
def most_recent_version(
    artifact: artifact_fs.FilesystemArtifact,
) -> artifact_fs.FilesystemArtifact:
    if not isinstance(artifact, LocalArtifact):
        raise ValueError(f"Artifact {artifact.name} is not a local artifact")

    if artifact._branch is not None and artifact._branch == "latest":
        return artifact

    possible_art = LocalArtifact(artifact.name, "latest")
    if possible_art._read_dirname != None:
        return possible_art

    obj_paths = sorted(
        pathlib.Path(artifact._root).iterdir(),
        key=os.path.getctime,
    )
    if not obj_paths:
        return None  # type: ignore

    latest_sym = None
    latest_dir = None

    for obj_path in reversed(obj_paths):
        if not obj_path.name.startswith(
            WORKING_DIR_PREFIX
        ) and not obj_path.name.startswith("."):
            if obj_path.is_symlink() and latest_sym is None:
                latest_sym = obj_path
            elif obj_path.is_dir() and latest_dir is None:
                latest_dir = obj_path
            if latest_sym is not None and latest_dir is not None:
                break

    if latest_dir is not None:
        if latest_sym is not None:
            latest_sym_points_to = latest_sym.resolve()
            if os.path.realpath(latest_sym_points_to) == os.path.realpath(latest_dir):
                return LocalArtifact(artifact.name, latest_sym.name)
            return LocalArtifact(artifact.name, obj_path.name)

    return None  # type: ignore


@op(
    name="FilesystemArtifact-previousURI",
)
def previous_uri(
    artifact: artifact_fs.FilesystemArtifact,
) -> typing.Optional[str]:
    return artifact.previous_uri()


@op(
    name="FilesystemArtifact-createdAt",
)
def created_at(
    artifact: artifact_fs.FilesystemArtifact,
) -> datetime.datetime:
    return artifact.created_at


@op(
    name="FilesystemArtifact-rootFromURI",
    pure=False,
)
def from_uri(
    uri: str,
) -> typing.Optional[artifact_fs.FilesystemArtifact]:
    ref = ref_base.Ref.from_str(uri)
    if not isinstance(ref, (artifact_fs.FilesystemArtifactRef)):
        return None
    return ref.artifact



import typing

import pytest

import weave
from weave.legacy.weave import context_state as _context
from weave.legacy.weave.weave_internal import make_const_node, make_output_node

_loading_builtins_token = _context.set_loading_built_ins()


@weave.op()
def no_arg_op() -> int:
    return 1


@weave.op()
def int_arg_op(a: int) -> int:
    return a + 1


@weave.op()
def int_args_op(a: int, b: int) -> int:
    return a + b


@weave.op()
def null_consuming_op(a: typing.Optional[int], b: int) -> int:
    if a == None:
        return b * 10
    return a + b  # type: ignore


_context.clear_loading_built_ins(_loading_builtins_token)


def test_basic_nullability():
    b = weave.save(2)
    maybe_int = weave.save([1, None])[0]
    null_int = weave.save([1, None])[1]

    assert weave.use(no_arg_op()) == 1

    assert weave.use(int_arg_op(b)) == 3
    assert weave.use(int_arg_op(maybe_int)) == 2
    assert weave.use(int_arg_op(null_int)) == None

    assert weave.use(int_args_op(b, b)) == 4
    assert weave.use(int_args_op(maybe_int, b)) == 3
    assert weave.use(int_args_op(null_int, b)) == None

    assert weave.use(null_consuming_op(b, b)) == 4
    assert weave.use(null_consuming_op(maybe_int, b)) == 3
    assert weave.use(null_consuming_op(null_int, b)) == 20


def test_basic_nullability_in_mappability():
    b_arr = weave.save([2])
    maybe_int_arr = weave.save(
        weave.legacy.weave.graph.ConstNode(
            weave.types.List(weave.types.optional(weave.types.Int())), [1, None]
        )
    )

    assert weave.use(int_arg_op(b_arr)) == [3]
    assert weave.use(int_arg_op(maybe_int_arr)) == [2, None]

    assert weave.use(int_args_op(b_arr, 2)) == [4]
    assert weave.use(int_args_op(maybe_int_arr, 2)) == [3, None]

    assert weave.use(null_consuming_op(b_arr, 2)) == [4]
    assert weave.use(null_consuming_op(maybe_int_arr, 2)) == [3, 20]


@pytest.mark.parametrize(
    "input_node, expected_type, expected",
    [
        # Base Case
        (
            lambda: weave.save({"a": 42}),
            weave.types.TypedDict({"a": weave.types.Int()}),
            42,
        ),
        # Maybe Case
        (
            lambda: weave.save([{"a": 42}, None])[0],
            weave.types.union(
                weave.types.NoneType(),
                weave.types.TypedDict({"a": weave.types.Int()}),
            ),
            42,
        ),
        # List of Maybe Case
        (
            lambda: weave.save([{"a": 42}, None]),
            weave.types.List(
                weave.types.union(
                    weave.types.NoneType(),
                    weave.types.TypedDict({"a": weave.types.Int()}),
                )
            ),
            [42, None],
        ),
        # Maybe list of Maybe Case
        (
            lambda: weave.save([[{"a": 42}, None], None])[0],
            weave.types.union(
                weave.types.NoneType(),
                weave.types.List(
                    weave.types.union(
                        weave.types.NoneType(),
                        weave.types.TypedDict({"a": weave.types.Int()}),
                    )
                ),
            ),
            [42, None],
        ),
    ],
)
def test_nullability_in_execution(input_node, expected_type, expected):
    input_node = input_node()
    assert input_node.type == expected_type
    # JS Weave0 Pick (noteable incorrect op name)
    js_pick = make_output_node(
        weave.types.Number(),
        "pick",
        {"obj": input_node, "key": make_const_node(weave.types.String(), "a")},
    )
    assert weave.use(js_pick) == expected

    # Ensure that this works at the type level so py dispatch works as well.
    py_pick = input_node.pick("a")
    assert weave.use(py_pick) == expected



import typing
from typing import Iterable

from weave.legacy.weave import stream_data_interfaces
from weave.legacy.weave.eager import WeaveIter
from weave.legacy.weave import artifact_wandb, uris
from weave.legacy.weave.run import Run


class RunStreamTableSpan:
    _attrs: stream_data_interfaces.TraceSpanDict

    def __init__(self, attrs: stream_data_interfaces.TraceSpanDict) -> None:
        self._attrs = attrs

    def __repr__(self) -> str:
        return f"<Run {self.uri} {self.id}>"

    @property
    def id(self) -> str:
        return self._attrs["span_id"]

    @property
    def trace_id(self) -> str:
        return self._attrs["trace_id"]

    @property
    def uri(self) -> str:
        return self._attrs["name"]

    @property
    def ui_url(self) -> str:
        raise NotImplementedError("ui_url not implemented")
        # gc = client_context.weave_client.require_weave_client()
        # return gc.run_ui_url(self)

    @property
    def op_ref(self) -> typing.Optional[artifact_wandb.WandbArtifactRef]:
        if self.uri.startswith("wandb-artifact"):
            parsed = uris.WeaveURI.parse(self.uri).to_ref()
            if not isinstance(parsed, artifact_wandb.WandbArtifactRef):
                raise ValueError("expected wandb artifact ref")
            return parsed
        return None

    @property
    def status_code(self) -> str:
        return self._attrs["status_code"]

    @property
    def attributes(self) -> dict[str, typing.Any]:
        attrs_dict = self._attrs.get("attributes", {})
        if not isinstance(attrs_dict, dict):
            return {}
        return {k: v for k, v in attrs_dict.items() if v != None}

    @property
    def inputs(self) -> dict[str, typing.Any]:
        input_dict = self._attrs.get("input", {})
        if not isinstance(input_dict, dict):
            return {}
        keys = input_dict.get("_keys")
        if keys is None:
            keys = [k for k in input_dict.keys() if input_dict[k] != None]
        return {k: input_dict[k] for k in keys}

    @property
    def output(self) -> typing.Any:
        if self.status_code != "SUCCESS":
            return None
        out = self._attrs.get("output")
        if isinstance(out, dict):
            keys = out.get("_keys")
            if keys is None:
                keys = [k for k in out.keys() if out[k] != None]
            output = {k: out.get(k) for k in keys}
            if "_result" in output:
                return output["_result"]
        return out

    @property
    def parent_id(self) -> typing.Optional[str]:
        return self._attrs["parent_id"]

    def parent(self) -> typing.Optional["Run"]:
        raise NotImplementedError("parent not implemented")
        # client = client_context.weave_client.require_weave_client()
        # if self.parent_id is None:
        #     return None
        # return client.run(self.parent_id)



import pytest
import wandb

import weave
from weave.legacy.weave import async_demo, compile, graph
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import use
from weave.legacy.weave.dispatch import RuntimeOutputNode
from weave.legacy.weave.ops_arrow import to_arrow
from weave.legacy.weave.ops_arrow.vectorize import raise_on_python_bailout
from weave.legacy.weave.wandb_interface.wandb_stream_table import StreamTable
from weave.legacy.weave.weave_internal import const, define_fn, make_const_node


def test_automatic_await_compile():
    twelve = async_demo.slowmult(3, 4, 0.01)
    twenty_four = async_demo.slowmult(2, twelve, 0.01)
    result = compile.compile([twenty_four])
    assert str(result[0]) == "2.slowmult(slowmult(3, 4, 0.01).await(), 0.01)"


@pytest.mark.parametrize(
    "nodes_with_expected_values",
    [
        # Simple graph with no modification needed
        [
            (
                3,
                graph.OutputNode(
                    types.Number(),
                    "add",
                    {
                        "lhs": graph.ConstNode(types.Number(), 1),
                        "rhs": graph.ConstNode(types.Number(), 2),
                    },
                ),
            )
        ],
        # Single Replacement Needed
        [
            (
                [3, 4, 5],
                graph.OutputNode(
                    types.Number(),
                    "add",
                    {
                        "lhs": graph.ConstNode(types.List(types.Number()), [1, 2, 3]),
                        "rhs": graph.ConstNode(types.Number(), 2),
                    },
                ),
            )
        ],
        # Nested Replacement Needed
        [
            (
                [13, 14, 15],
                graph.OutputNode(
                    types.Number(),
                    "add",
                    {
                        "lhs": graph.OutputNode(
                            types.Number(),
                            "add",
                            {
                                "lhs": graph.ConstNode(
                                    types.List(types.Number()), [1, 2, 3]
                                ),
                                "rhs": graph.ConstNode(types.Number(), 2),
                            },
                        ),
                        "rhs": graph.ConstNode(types.Number(), 10),
                    },
                ),
            )
        ],
        # Nested Replacement Through Valid Node
        [
            (
                [13, 14, 15],
                graph.OutputNode(
                    types.Number(),
                    "add",
                    {
                        "lhs": graph.OutputNode(
                            types.Number(),
                            "index",
                            {
                                "arr": graph.OutputNode(
                                    types.List(types.Number()),
                                    "list",
                                    {
                                        "0": graph.OutputNode(
                                            types.Number(),
                                            "add",
                                            {
                                                "lhs": graph.ConstNode(
                                                    types.List(types.Number()),
                                                    [1, 2, 3],
                                                ),
                                                "rhs": graph.ConstNode(
                                                    types.Number(), 2
                                                ),
                                            },
                                        ),
                                    },
                                ),
                                "index": graph.ConstNode(types.Number(), 0),
                            },
                        ),
                        "rhs": graph.ConstNode(types.Number(), 10),
                    },
                ),
            )
        ],
    ],
)
def test_executing_js_graphs(nodes_with_expected_values):
    nodes = [node for exp_val, node in nodes_with_expected_values]
    exp_vals = [exp_val for exp_val, node in nodes_with_expected_values]
    assert use(nodes) == exp_vals


def test_executing_js_multi_root():
    node = graph.OutputNode(
        types.Number(),
        "add",
        {
            "lhs": graph.ConstNode(
                types.List(types.Number()),
                [1, 2, 3],
            ),
            "rhs": graph.ConstNode(types.Number(), 2),
        },
    )
    node2 = graph.OutputNode(
        types.Number(),
        "add",
        {
            "lhs": node,
            "rhs": graph.ConstNode(types.Number(), 4),
        },
    )
    assert use([node, node2]) == [[3, 4, 5], [7, 8, 9]]


def test_optimize_merge_empty_dict():
    non_empty_dict = weave.legacy.weave.ops.dict_(a=5, b=2)
    assert (
        compile.compile_simple_optimizations(
            [weave.legacy.weave.ops.TypedDict.merge(non_empty_dict, weave.legacy.weave.ops.dict_())]
        )[0].to_json()
        == non_empty_dict.to_json()
    )
    assert (
        compile.compile_simple_optimizations(
            [weave.legacy.weave.ops.TypedDict.merge(weave.legacy.weave.ops.dict_(), non_empty_dict)]
        )[0].to_json()
        == non_empty_dict.to_json()
    )
    non_simplified_merge = weave.legacy.weave.ops.TypedDict.merge(
        weave.legacy.weave.ops.dict_(j=3), non_empty_dict
    )
    assert (
        compile.compile_simple_optimizations([non_simplified_merge])[0].to_json()
        == non_simplified_merge.to_json()
    )


def test_compile_lambda_uniqueness():
    list_node_1 = weave.legacy.weave.ops.make_list(a=make_const_node(weave.types.Number(), 1))
    list_node_2 = weave.legacy.weave.ops.make_list(a=make_const_node(weave.types.Number(), 2))
    fn_node = define_fn({"row": weave.types.Number()}, lambda row: row + 1)
    mapped_1 = list_node_1.map(fn_node)
    mapped_2 = list_node_2.map(fn_node)
    combined = weave.legacy.weave.ops.make_list(a=mapped_1, b=mapped_2)
    concatted = combined.concat()

    # list node contains 2 nodes (const, list), x 2 = 4
    # fn node contains 3 nodes (row, add, const) + the fn node itself, x1 = 4
    # map node contains 1 node, x2 = 2
    # combined node contains 1 node, x1 = 1
    # concat node contains 1 node, x1 = 1
    # total = 12
    assert graph.count(concatted) == 12

    # However, after lambda compilation, we should get
    # 3 more nodes (new row, add, and fn), the const 1
    # is not deduped because in one case (inside the lambda function),
    # it is an int and in the other, it is a number
    compiled = compile.compile([concatted])[0]

    assert graph.count(compiled) == 15


# We actually don't want this to work because it would require
# mutating a static lambda function!
# def test_compile_through_execution(user_by_api_key_in_env):
#     run = wandb.init(project="project_exists")
#     for i in range(10):
#         run.log({"val": i, "cat": i % 2})
#     run.finish()

#     """
#     This test demonstrates successful execution when there is an explicit
#     const function instead of a direct node (resulting in an intermediate execution op)
#     """
#     history_node = weave.legacy.weave.ops.project(run.entity, run.project).run(run.id).history2()
#     pick = const(history_node).pick("val")
#     res = weave.use(pick)
#     assert res.to_pylist_notags() == list(range(10))


def test_compile_through_function_call(user_by_api_key_in_env):
    run = wandb.init(project="project_exists")
    for i in range(10):
        run.log({"val": i, "cat": i % 2})
    run.finish()

    """
    This test demonstrates successful execution when there is an explicit
    function-__call__ in the graph)
    """
    fn_node = define_fn(
        {"entity_name": types.String()},
        lambda entity_name: (
            weave.legacy.weave.ops.project(entity_name, run.project).run(run.id).history2()
        ),
    )
    called_node = fn_node(run.entity)
    pick = called_node.pick("val")
    res = weave.use(pick)
    assert res.to_pylist_notags() == list(range(10))


def test_compile_list_flatten_to_awl_concat():
    # 4 cases: list of lists, list of awls, awl of lists, awl of awls
    # When the outer list-structure is a list, we want to dispatch to concat, preferably AWL-concat
    # when the outer list-structure is an AWL, we want to dispatch ensure that we use AWL ops
    # list of lists
    list_list_node = weave.legacy.weave.ops.make_list(a=[1], b=[2])
    list_list_node_concat = list_list_node.concat()
    list_list_node_flatten = list_list_node.flatten()
    list_list_node_concat_compiled = compile.compile([list_list_node_concat])[0]
    list_list_node_flatten_compiled = compile.compile([list_list_node_flatten])[0]
    assert list_list_node_concat_compiled.from_op.name == "concat"
    assert list_list_node_flatten_compiled.from_op.name == "flatten"
    # list of awls
    list_awl_node = weave.legacy.weave.ops.make_list(a=to_arrow([1]), b=to_arrow([2]))
    list_awl_node_concat = list_awl_node.concat()
    list_awl_node_flatten = list_awl_node.flatten()
    list_awl_node_concat_compiled = compile.compile([list_awl_node_concat])[0]
    list_awl_node_flatten_compiled = compile.compile([list_awl_node_flatten])[0]
    assert list_awl_node_concat_compiled.from_op.name == "ArrowWeaveList-concat"
    # **THIS IS THE SPECIAL CASE 1**: the flatten operation is transformed into a concat!
    assert list_awl_node_flatten_compiled.from_op.name == "ArrowWeaveList-concat"
    # awl of lists
    awl_list_node = weave.save(to_arrow([[1], [2]]))
    awl_list_node_concat = awl_list_node.concat()
    awl_list_node_flatten = awl_list_node.flatten()
    awl_list_node_concat_compiled = compile.compile([awl_list_node_concat])[0]
    awl_list_node_flatten_compiled = compile.compile([awl_list_node_flatten])[0]
    # **THIS IS THE SPECIAL CASE 2**: the concat operation is transformed into a flatten!
    assert awl_list_node_concat_compiled.from_op.name == "ArrowWeaveList-flatten"
    assert awl_list_node_flatten_compiled.from_op.name == "ArrowWeaveList-flatten"
    # awl of awls
    awl_awl_node = weave.save(to_arrow([to_arrow([1]), to_arrow([2])]))
    awl_awl_node_concat = awl_awl_node.concat()
    awl_awl_node_flatten = awl_awl_node.flatten()
    awl_awl_node_concat_compiled = compile.compile([awl_awl_node_concat])[0]
    awl_awl_node_flatten_compiled = compile.compile([awl_awl_node_flatten])[0]
    assert awl_awl_node_concat_compiled.from_op.name == "ArrowWeaveList-concat"
    assert awl_awl_node_flatten_compiled.from_op.name == "ArrowWeaveList-flatten"

    results = weave.use(
        [
            list_list_node_concat,
            list_list_node_flatten,
            list_awl_node_concat,
            list_awl_node_flatten,
            awl_list_node_concat,
            awl_list_node_flatten,
            awl_awl_node_concat,
            awl_awl_node_flatten,
        ]
    )

    for result in results[:2]:
        assert result == [1, 2]
    for result in results[2:]:
        assert result.to_pylist_notags() == [1, 2]


def test_compile_lambda_on_refineable(user_by_api_key_in_env):
    st = StreamTable(
        "test_table",
        project_name="stream-tables",
        entity_name=user_by_api_key_in_env.username,
        _disable_async_file_stream=True,
    )

    for i in range(10):
        st.log({"num": i, "str": str(i)})

    st.finish()

    rows_node = st.rows()
    rows_node_copy = RuntimeOutputNode(
        types.List(types.TypedDict({})),
        rows_node.from_op.name,
        rows_node.from_op.inputs,
    )
    mapped_node = rows_node_copy.map(lambda row: row["num"] + 1)
    with raise_on_python_bailout():
        val = use(mapped_node)
    assert val.to_pylist_raw() == list(range(1, 11))



# Clickhouse Trace Server Manager
import logging
import os
import typing

from clickhouse_connect.driver.client import Client as CHClient

logger = logging.getLogger(__name__)


class ClickHouseTraceServerMigrator:
    ch_client: CHClient

    def __init__(
        self,
        ch_client: CHClient,
    ):
        super().__init__()
        self.ch_client = ch_client
        self._initialize_migration_db()

    def apply_migrations(
        self, target_db: str, target_version: typing.Optional[int] = None
    ) -> None:
        status = self._get_migration_status(target_db)
        logger.info(f"""`{target_db}` migration status: {status}""")
        if status["partially_applied_version"]:
            logger.info(
                f"Unable to apply migrations to `{target_db}`. Found partially applied migration version {status['partially_applied_version']}. Please fix the database manually and try again."
            )
            return
        migration_map = self._get_migrations()
        migrations_to_apply = self._determine_migrations_to_apply(
            status["curr_version"], migration_map, target_version
        )
        if len(migrations_to_apply) == 0:
            logger.info(f"No migrations to apply to `{target_db}`")
            return
        logger.info(f"Migrations to apply: {migrations_to_apply}")
        if status["curr_version"] == 0:
            self.ch_client.command(f"CREATE DATABASE IF NOT EXISTS {target_db}")
        for target_version, migration_file in migrations_to_apply:
            self._apply_migration(target_db, target_version, migration_file)

    def _initialize_migration_db(self) -> None:
        self.ch_client.command(
            """
            CREATE DATABASE IF NOT EXISTS db_management
        """
        )
        self.ch_client.command(
            """
            CREATE TABLE IF NOT EXISTS db_management.migrations
            (
                db_name String,
                curr_version UInt64,
                partially_applied_version UInt64 NULL,
            )
            ENGINE = MergeTree()
            ORDER BY (db_name)
        """
        )

    def _get_migration_status(self, db_name: str) -> typing.Dict:
        column_names = ["db_name", "curr_version", "partially_applied_version"]
        select_columns = ", ".join(column_names)
        query = f"""
            SELECT {select_columns} FROM db_management.migrations WHERE db_name = '{db_name}'
        """
        res = self.ch_client.query(query)
        result_rows = res.result_rows
        if res is None or len(result_rows) == 0:
            self.ch_client.insert(
                "db_management.migrations",
                data=[[db_name, 0, None]],
                column_names=column_names,
            )
        res = self.ch_client.query(query)
        result_rows = res.result_rows
        if res is None or len(result_rows) == 0:
            raise Exception("Migration table not found")

        return dict(zip(column_names, result_rows[0]))

    def _get_migrations(
        self,
    ) -> typing.Dict[int, typing.Dict[str, typing.Optional[str]]]:
        migration_dir = os.path.join(os.path.dirname(__file__), "migrations")
        migration_files = os.listdir(migration_dir)
        migration_map: typing.Dict[int, typing.Dict[str, typing.Optional[str]]] = {}
        max_version = 0
        for file in migration_files:
            if not file.endswith(".up.sql") and not file.endswith(".down.sql"):
                raise Exception(f"Invalid migration file: {file}")
            file_name_parts = file.split("_", 1)
            if len(file_name_parts) <= 1:
                raise Exception(f"Invalid migration file: {file}")
            version = int(file_name_parts[0], 10)
            if version < 1:
                raise Exception(f"Invalid migration file: {file}")

            is_up = file.endswith(".up.sql")

            if version not in migration_map:
                migration_map[version] = {"up": None, "down": None}

            if is_up:
                if migration_map[version]["up"] is not None:
                    raise Exception(f"Duplicate migration file for version {version}")
                migration_map[version]["up"] = file
            else:
                if migration_map[version]["down"] is not None:
                    raise Exception(f"Duplicate migration file for version {version}")
                migration_map[version]["down"] = file

            if version > max_version:
                max_version = version

        if len(migration_map) == 0:
            raise Exception("No migrations found")

        if max_version != len(migration_map):
            raise Exception(
                f"Invalid migration versioning. Expected {max_version} migrations but found {len(migration_map)}"
            )

        for version in range(1, max_version + 1):
            if version not in migration_map:
                raise Exception(f"Missing migration file for version {version}")
            if migration_map[version]["up"] is None:
                raise Exception(f"Missing up migration file for version {version}")
            if migration_map[version]["down"] is None:
                raise Exception(f"Missing down migration file for version {version}")

        return migration_map

    def _determine_migrations_to_apply(
        self,
        current_version: int,
        migration_map: typing.Dict,
        target_version: typing.Optional[int] = None,
    ) -> typing.List[typing.Tuple[int, str]]:
        if target_version is None:
            target_version = len(migration_map)
        if target_version < 0 or target_version > len(migration_map):
            raise Exception(f"Invalid target version: {target_version}")

        if target_version > current_version:
            res = []
            for i in range(current_version + 1, target_version + 1):
                if migration_map[i]["up"] is None:
                    raise Exception(f"Missing up migration file for version {i}")
                res.append((i, f"{migration_map[i]['up']}"))
            return res
        if target_version < current_version:
            res = []
            for i in range(current_version, target_version, -1):
                if migration_map[i]["down"] is None:
                    raise Exception(f"Missing down migration file for version {i}")
                res.append((i - 1, f"{migration_map[i]['down']}"))
            return res

        return []

    def _apply_migration(
        self, target_db: str, target_version: int, migration_file: str
    ) -> None:
        logger.info(f"Applying migration {migration_file} to `{target_db}`")
        migration_dir = os.path.join(os.path.dirname(__file__), "migrations")
        migration_file_path = os.path.join(migration_dir, migration_file)
        with open(migration_file_path, "r") as f:
            migration_sql = f.read()
        self.ch_client.command(
            f"""
            ALTER TABLE db_management.migrations UPDATE partially_applied_version = {target_version} WHERE db_name = '{target_db}'
        """
        )
        migration_sub_commands = migration_sql.split(";")
        for command in migration_sub_commands:
            command = command.strip()
            if len(command) == 0:
                continue
            curr_db = self.ch_client.database
            self.ch_client.database = target_db
            self.ch_client.command(command)
            self.ch_client.database = curr_db
        self.ch_client.command(
            f"""
            ALTER TABLE db_management.migrations UPDATE curr_version = {target_version}, partially_applied_version = NULL WHERE db_name = '{target_db}'
        """
        )
        logger.info(f"Migration {migration_file} applied to `{target_db}`")



import uuid
import gradio as gr

import PIL
import dataclasses


@dataclasses.dataclass
class Prediction:
    logits: dict[str, float]
    prediction_id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))


class PredictionServiceInterface:
    def predict(self, pil_image: "PIL.Image") -> Prediction:  # type: ignore
        raise NotImplementedError

    def record_feedback(self, prediction_id: str, feedback: int) -> None:
        return None


def render_app(service: PredictionServiceInterface):
    def on_predict(pil_image):
        if pil_image is None:
            return None
        prediction = service.predict(pil_image)
        return prediction.logits, prediction

    def make_feedback_callback(feedback):
        def on_feedback(current_prediction):
            if current_prediction is None:
                return
            service.record_feedback(current_prediction.prediction_id, feedback)

        return on_feedback

    with gr.Blocks() as demo:
        current_prediction = gr.State(value=None)
        with gr.Column():
            with gr.Row():
                inp = gr.Sketchpad(
                    label="Draw a number between 0 and 9",
                    brush_radius=5,
                    type="pil",
                    shape=(128, 128),
                )
                out = gr.Label(label="Prediction", num_top_classes=4)
            btn = gr.Button("Predict")
            with gr.Row():
                btns = [
                    gr.ClearButton(components=[inp, out], value=str(i))
                    for i in range(10)
                ]

        btn.click(fn=on_predict, inputs=inp, outputs=[out, current_prediction])
        for btn_ndx, fb_btn in enumerate(btns):
            fb_btn.click(
                fn=make_feedback_callback(btn_ndx),
                inputs=current_prediction,
                outputs=[],
            )

    demo.launch(quiet=True, show_api=False, height=550)



import typing

from weave.legacy.weave import decorator_op, graph, graph_debug


class RowType(typing.TypedDict):
    a: int
    b: int
    c: int


@decorator_op.op()
def _test_cn_op1() -> list[int]:
    return [1, 2, 3]


@decorator_op.op()
def _test_cn_op2(v: int, addend: int) -> RowType:
    return {"a": v, "b": v + addend, "c": v + addend * 2}


def test_combine_nodes():
    common_node = _test_cn_op1().map(lambda x: _test_cn_op2(x, 1))
    n1 = common_node[0]["a"]
    n2 = common_node[0]["b"]
    n3 = common_node[0]["c"]
    n4 = common_node[1]["a"]
    n5 = common_node[1]["b"]

    result = graph_debug.combine_common_nodes([n1, n2, n3, n4, n5])
    assert len(result) == 1
    assert (
        graph.node_expr_str(result[0])
        == "_test_cn_op1().map(_test_cn_op2(row, 1))[EACH[0, 1]][EACH['a', 'b', 'c']]"
    )


def test_to_assignment_form():
    common_node = _test_cn_op1().map(lambda x: _test_cn_op2(x, 1))
    n1 = common_node[0]["a"]
    n2 = common_node[0]["b"]
    n3 = common_node[0]["c"]
    n4 = common_node[1]["a"]
    n5 = common_node[1]["b"]

    result = graph_debug.to_assignment_form([n1, n2, n3, n4, n5])
    print(graph_debug.assignments_string(result))
    assert (
        graph_debug.assignments_string(result)
        == """var0 = op-_test_cn_op1()
  .map(  op-_test_cn_op2(row, 1)) 
var1 = list-__getitem__(var0, 1)
  .typedDict-pick(b) 
var2 = list-__getitem__(var0, 1)
  .typedDict-pick(a) 
var3 = list-__getitem__(var0, 0)
  .typedDict-pick(c) 
var4 = list-__getitem__(var0, 0)
  .typedDict-pick(b) 
var5 = list-__getitem__(var0, 0)
  .typedDict-pick(a) """  # noqa: W291
    )



import os

import cohere
import pytest

import weave
from weave.integrations.integration_utilities import _get_call_output, op_name_from_ref

cohere_model = "command"  # You can change this to a specific model if needed


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
def test_cohere(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("COHERE_API_KEY", "DUMMY_API_KEY")

    cohere_client = cohere.Client(api_key=api_key)

    response = cohere_client.chat(
        model=cohere_model,
        message="Hello, Cohere!",
        max_tokens=1024,
    )

    exp = response.text
    assert exp.strip() != ""
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    assert call.started_at < call.ended_at
    assert op_name_from_ref(call.op_name) == "cohere.Client.chat"
    output = call.output
    assert output.text == exp
    assert output.generation_id == response.generation_id
    assert output.citations == response.citations
    assert output.documents == response.documents
    assert output.is_search_required == response.is_search_required
    assert output.search_queries == response.search_queries
    assert output.search_results == response.search_results
    assert (
        output.meta.billed_units.input_tokens == response.meta.billed_units.input_tokens
    )
    assert (
        output.meta.billed_units.output_tokens
        == response.meta.billed_units.output_tokens
    )
    assert (
        output.meta.billed_units.search_units == response.meta.billed_units.search_units
    )
    assert (
        output.meta.billed_units.classifications
        == response.meta.billed_units.classifications
    )
    assert output.meta.tokens.input_tokens == response.meta.tokens.input_tokens
    assert output.meta.tokens.output_tokens == response.meta.tokens.output_tokens


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
def test_cohere_stream(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("COHERE_API_KEY", "DUMMY_API_KEY")
    cohere_client = cohere.Client(api_key=api_key)

    stream = cohere_client.chat_stream(
        model=cohere_model,
        message="Hello, Cohere!",
        max_tokens=1024,
    )

    # they accumulate for us in the last message
    for event in stream:
        pass

    response = event.response  # the NonStreamedChatResponse
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    assert call.started_at < call.ended_at
    assert op_name_from_ref(call.op_name) == "cohere.Client.chat_stream"
    output = call.output
    assert output.text == response.text
    assert output.generation_id == response.generation_id
    summary = call.summary
    assert summary is not None
    assert output.generation_id == response.generation_id
    assert output.citations == response.citations
    assert output.documents == response.documents
    assert output.is_search_required == response.is_search_required
    assert output.search_queries == response.search_queries
    assert output.search_results == response.search_results
    assert (
        output.meta.billed_units.input_tokens == response.meta.billed_units.input_tokens
    )
    assert (
        output.meta.billed_units.output_tokens
        == response.meta.billed_units.output_tokens
    )
    assert (
        output.meta.billed_units.search_units == response.meta.billed_units.search_units
    )
    assert (
        output.meta.billed_units.classifications
        == response.meta.billed_units.classifications
    )
    assert output.meta.tokens.input_tokens == response.meta.tokens.input_tokens
    assert output.meta.tokens.output_tokens == response.meta.tokens.output_tokens


@pytest.mark.asyncio
@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
async def test_cohere_async(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("COHERE_API_KEY", "DUMMY_API_KEY")

    cohere_client = cohere.AsyncClient(api_key=api_key)

    response = await cohere_client.chat(
        model=cohere_model,
        message="Hello, Async Cohere!",
        max_tokens=1024,
    )

    exp = response.text
    assert exp.strip() != ""
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    assert call.started_at < call.ended_at
    assert op_name_from_ref(call.op_name) == "cohere.AsyncClient.chat"
    output = call.output
    assert output.text == exp
    assert output.generation_id == response.generation_id
    assert output.citations == response.citations
    assert output.documents == response.documents
    assert output.is_search_required == response.is_search_required
    assert output.search_queries == response.search_queries
    assert output.search_results == response.search_results
    assert (
        output.meta.billed_units.input_tokens == response.meta.billed_units.input_tokens
    )
    assert (
        output.meta.billed_units.output_tokens
        == response.meta.billed_units.output_tokens
    )
    assert (
        output.meta.billed_units.search_units == response.meta.billed_units.search_units
    )
    assert (
        output.meta.billed_units.classifications
        == response.meta.billed_units.classifications
    )
    assert output.meta.tokens.input_tokens == response.meta.tokens.input_tokens
    assert output.meta.tokens.output_tokens == response.meta.tokens.output_tokens


@pytest.mark.asyncio
@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
async def test_cohere_async_stream(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("COHERE_API_KEY", "DUMMY_API_KEY")
    cohere_client = cohere.AsyncClient(api_key=api_key)

    stream = cohere_client.chat_stream(
        model=cohere_model,
        message="Hello, Async Cohere Stream!",
        max_tokens=1024,
    )

    async for event in stream:
        pass

    response = event.response  # the NonStreamedChatResponse
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    assert call.started_at < call.ended_at
    assert op_name_from_ref(call.op_name) == "cohere.AsyncClient.chat_stream"
    output = call.output
    assert output.text == response.text
    assert output.generation_id == response.generation_id
    summary = call.summary
    assert summary is not None
    assert output.generation_id == response.generation_id
    assert output.citations == response.citations
    assert output.documents == response.documents
    assert output.is_search_required == response.is_search_required
    assert output.search_queries == response.search_queries
    assert output.search_results == response.search_results
    assert (
        output.meta.billed_units.input_tokens == response.meta.billed_units.input_tokens
    )
    assert (
        output.meta.billed_units.output_tokens
        == response.meta.billed_units.output_tokens
    )
    assert (
        output.meta.billed_units.search_units == response.meta.billed_units.search_units
    )
    assert (
        output.meta.billed_units.classifications
        == response.meta.billed_units.classifications
    )
    assert output.meta.tokens.input_tokens == response.meta.tokens.input_tokens
    assert output.meta.tokens.output_tokens == response.meta.tokens.output_tokens


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
def test_cohere_v2(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("COHERE_API_KEY", "DUMMY_API_KEY")

    cohere_client = cohere.ClientV2(api_key=api_key)

    response = cohere_client.chat(
        model=cohere_model,
        messages=[{"role": "user", "content": "count to three"}],
        max_tokens=1024,
    )
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    assert call.started_at < call.ended_at
    assert op_name_from_ref(call.op_name) == "cohere.ClientV2.chat"
    output = _get_call_output(call)

    assert output.message.content[0].text == response.message.content[0].text
    assert output.id == response.id
    assert output.finish_reason == response.finish_reason
    assert output.message.role == response.message.role
    assert output.message.tool_calls == response.message.tool_calls
    assert output.message.tool_plan == response.message.tool_plan
    assert output.message.citations == response.message.citations

    assert (
        output.usage.billed_units.input_tokens
        == response.meta["billed_units"]["input_tokens"]
    )
    assert (
        output.usage.billed_units.output_tokens
        == response.meta["billed_units"]["output_tokens"]
    )
    assert output.usage.tokens.input_tokens == response.meta["tokens"]["input_tokens"]
    assert output.usage.tokens.output_tokens == response.meta["tokens"]["output_tokens"]


@pytest.mark.asyncio
@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
async def test_cohere_async_v2(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("COHERE_API_KEY", "DUMMY_API_KEY")

    cohere_client = cohere.AsyncClientV2(api_key=api_key)

    response = await cohere_client.chat(
        model=cohere_model,
        messages=[{"role": "user", "content": "count to three"}],
        max_tokens=1024,
    )
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    assert call.started_at < call.ended_at
    assert op_name_from_ref(call.op_name) == "cohere.AsyncClientV2.chat"
    output = _get_call_output(call)

    assert output.message.content[0].text == response.message.content[0].text
    assert output.id == response.id
    assert output.finish_reason == response.finish_reason
    assert output.message.role == response.message.role
    assert output.message.tool_calls == response.message.tool_calls
    assert output.message.tool_plan == response.message.tool_plan
    assert output.message.citations == response.message.citations

    assert (
        output.usage.billed_units.input_tokens
        == response.meta["billed_units"]["input_tokens"]
    )
    assert (
        output.usage.billed_units.output_tokens
        == response.meta["billed_units"]["output_tokens"]
    )
    assert output.usage.tokens.input_tokens == response.meta["tokens"]["input_tokens"]
    assert output.usage.tokens.output_tokens == response.meta["tokens"]["output_tokens"]


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
def test_cohere_stream_v2(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("COHERE_API_KEY", "DUMMY_API_KEY")

    cohere_client = cohere.ClientV2(api_key=api_key)

    response = cohere_client.chat_stream(
        model=cohere_model,
        messages=[{"role": "user", "content": "count to three"}],
        max_tokens=10,
    )

    all_content = ""
    for event in response:
        if event is not None:
            if event.type == "message-start":
                id = event.id
                role = event.delta.message.role
            if event.type == "content-delta":
                all_content += event.delta.message.content.text
            if event.type == "message-end":
                finish_reason = event.delta.finish_reason

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    assert call.started_at < call.ended_at
    assert op_name_from_ref(call.op_name) == "cohere.ClientV2.chat_stream"
    output = _get_call_output(call)

    assert output.message.content[0] == all_content
    assert output.id == id
    assert output.finish_reason == finish_reason
    assert output.message.role == role

    assert output.usage.billed_units.input_tokens == 3.0
    assert output.usage.billed_units.output_tokens == 9.0
    assert output.usage.tokens.input_tokens == 65.0
    assert output.usage.tokens.output_tokens == 10.0


@pytest.mark.asyncio
@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
async def test_cohere_async_stream_v2(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("COHERE_API_KEY", "DUMMY_API_KEY")

    cohere_client = cohere.AsyncClientV2(api_key=api_key)

    response = cohere_client.chat_stream(
        model=cohere_model,
        messages=[{"role": "user", "content": "count to three"}],
        max_tokens=15,
    )

    all_content = ""
    async for event in response:
        if event is not None:
            if event.type == "message-start":
                id = event.id
                role = event.delta.message.role
            if event.type == "content-delta":
                all_content += event.delta.message.content.text
            if event.type == "message-end":
                finish_reason = event.delta.finish_reason

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    assert call.started_at < call.ended_at
    assert op_name_from_ref(call.op_name) == "cohere.AsyncClientV2.chat_stream"
    output = _get_call_output(call)

    assert output.message.content[0] == all_content
    assert output.id == id
    assert output.finish_reason == finish_reason
    assert output.message.role == role

    assert output.usage.billed_units.input_tokens == 3.0
    assert output.usage.billed_units.output_tokens == 15.0
    assert output.usage.tokens.input_tokens == 65.0
    assert output.usage.tokens.output_tokens == 15.0
